# type: ignore
import json
import os
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any, Optional

from pydantic import Field, Secret, ValidationError
from pydantic.functional_validators import BeforeValidator

from unstructured_ingest.embed.interfaces import (
    AsyncBaseEmbeddingEncoder,
    BaseEmbeddingEncoder,
    EmbeddingConfig,
)
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.errors import UserAuthError

if TYPE_CHECKING:
    from vertexai.language_models import TextEmbeddingModel


def conform_string_to_dict(value: Any) -> dict:
    if isinstance(value, dict):
        return value
    if isinstance(value, str):
        return json.loads(value)
    raise ValidationError(f"Input could not be mapped to a valid dict: {value}")


ApiKeyType = Secret[Annotated[dict, BeforeValidator(conform_string_to_dict)]]


class VertexAIEmbeddingConfig(EmbeddingConfig):
    api_key: ApiKeyType
    embedder_model_name: Optional[str] = Field(
        default="textembedding-gecko@001", alias="model_name"
    )

    def wrap_error(self, e: Exception) -> Exception:
        from google.auth.exceptions import GoogleAuthError

        if isinstance(e, GoogleAuthError):
            return UserAuthError(e)
        return e

    def register_application_credentials(self):
        # TODO look into passing credentials in directly, rather than via env var and tmp file
        application_credentials_path = Path("/tmp") / "google-vertex-app-credentials.json"
        with application_credentials_path.open("w+") as credentials_file:
            json.dump(self.api_key.get_secret_value(), credentials_file)
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = str(application_credentials_path)

    @requires_dependencies(
        ["vertexai"],
        extras="embed-vertexai",
    )
    def get_client(self) -> "TextEmbeddingModel":
        """Creates a VertexAI python client to embed elements."""
        from vertexai.language_models import TextEmbeddingModel

        self.register_application_credentials()
        return TextEmbeddingModel.from_pretrained(self.embedder_model_name)


@dataclass
class VertexAIEmbeddingEncoder(BaseEmbeddingEncoder):
    config: VertexAIEmbeddingConfig

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    def embed_query(self, query):
        return self._embed_documents(elements=[query])[0]

    def embed_documents(self, elements: list[dict]) -> list[dict]:
        embeddings = self._embed_documents([e.get("text", "") for e in elements])
        elements_with_embeddings = self._add_embeddings_to_elements(elements, embeddings)
        return elements_with_embeddings

    @requires_dependencies(
        ["vertexai"],
        extras="embed-vertexai",
    )
    def _embed_documents(self, elements: list[str]) -> list[list[float]]:
        from vertexai.language_models import TextEmbeddingInput

        inputs = [TextEmbeddingInput(text=element) for element in elements]
        try:
            client = self.config.get_client()
            embeddings = client.get_embeddings(inputs)
        except Exception as e:
            raise self.wrap_error(e=e)
        return [e.values for e in embeddings]


@dataclass
class AsyncVertexAIEmbeddingEncoder(AsyncBaseEmbeddingEncoder):
    config: VertexAIEmbeddingConfig

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    async def embed_query(self, query):
        embedding = await self._embed_documents(elements=[query])
        return embedding[0]

    async def embed_documents(self, elements: list[dict]) -> list[dict]:
        embeddings = await self._embed_documents([e.get("text", "") for e in elements])
        elements_with_embeddings = self._add_embeddings_to_elements(elements, embeddings)
        return elements_with_embeddings

    @requires_dependencies(
        ["vertexai"],
        extras="embed-vertexai",
    )
    async def _embed_documents(self, elements: list[str]) -> list[list[float]]:
        from vertexai.language_models import TextEmbeddingInput

        inputs = [TextEmbeddingInput(text=element) for element in elements]
        try:
            client = self.config.get_client()
            embeddings = await client.get_embeddings_async(inputs)
        except Exception as e:
            raise self.wrap_error(e=e)
        return [e.values for e in embeddings]



import random
from typing import Any

import faker
import pytest

from test.unit.v2.utils.data_generator import generate_random_dictionary
from unstructured_ingest.embed.huggingface import (
    HuggingFaceEmbeddingConfig,
    HuggingFaceEmbeddingEncoder,
)

fake = faker.Faker()


def generate_embedder_config_params() -> dict:
    params = {}
    if random.random() < 0.5:
        params["embed_model_name"] = fake.word() if random.random() < 0.5 else None
        params["embedder_model_kwargs"] = (
            generate_random_dictionary(key_type=str, value_type=Any)
            if random.random() < 0.5
            else None
        )
        params["encode_kwargs"] = (
            generate_random_dictionary(key_type=str, value_type=Any)
            if random.random() < 0.5
            else None
        )
        params["cache_folder"] = fake.file_path() if random.random() < 0.5 else None
    return params


@pytest.mark.parametrize(
    "embedder_config_params", [generate_embedder_config_params() for i in range(10)]
)
def test_embedder_config(embedder_config_params: dict):
    embedder_config = HuggingFaceEmbeddingConfig.model_validate(embedder_config_params)
    assert embedder_config


@pytest.mark.parametrize(
    "embedder_config_params", [generate_embedder_config_params() for i in range(10)]
)
def test_embedder(embedder_config_params: dict):
    embedder_config = HuggingFaceEmbeddingConfig.model_validate(embedder_config_params)
    embedder = HuggingFaceEmbeddingEncoder(config=embedder_config)
    assert embedder



from unstructured_ingest.embed.vertexai import VertexAIEmbeddingConfig, VertexAIEmbeddingEncoder


def test_embed_documents_does_not_break_element_to_dict(mocker):
    # Mocked client with the desired behavior for embed_documents
    mock_responses = []
    for i in [1, 2]:
        mock_response = mocker.Mock()
        mocker.patch.object(mock_response, "values", i)
        mock_responses.append(mock_response)

    mock_client = mocker.MagicMock()
    mock_client.get_embeddings.return_value = mock_responses

    # Mock create_client to return our mock_client
    mocker.patch.object(VertexAIEmbeddingConfig, "get_client", return_value=mock_client)
    encoder = VertexAIEmbeddingEncoder(config=VertexAIEmbeddingConfig(api_key={"api_key": "value"}))
    raw_elements = [{"text": f"This is sentence {i + 1}"} for i in range(2)]

    elements = encoder.embed_documents(
        elements=raw_elements,
    )
    assert len(elements) == 2
    assert elements[0]["text"] == "This is sentence 1"
    assert elements[1]["text"] == "This is sentence 2"



import json
import time
from pathlib import Path

import docker
import pytest
from _pytest.fixtures import TopRequest
from pymilvus import (
    CollectionSchema,
    DataType,
    FieldSchema,
    MilvusClient,
)
from pymilvus.milvus_client import IndexParams

from test.integration.connectors.utils.constants import (
    DESTINATION_TAG,
    VECTOR_DB_TAG,
    env_setup_path,
)
from test.integration.connectors.utils.docker import healthcheck_wait
from test.integration.connectors.utils.docker_compose import docker_compose_context
from test.integration.connectors.utils.validation.destination import (
    StagerValidationConfigs,
    stager_validation,
)
from unstructured_ingest.error import DestinationConnectionError
from unstructured_ingest.v2.interfaces import FileData, SourceIdentifiers
from unstructured_ingest.v2.processes.connectors.milvus import (
    CONNECTOR_TYPE,
    MilvusConnectionConfig,
    MilvusUploader,
    MilvusUploaderConfig,
    MilvusUploadStager,
)

DB_NAME = "test_database"
EXISTENT_COLLECTION_NAME = "test_collection"
NONEXISTENT_COLLECTION_NAME = "nonexistent_collection"
DB_URI = "http://localhost:19530"


def get_schema() -> CollectionSchema:
    id_field = FieldSchema(
        name="id", dtype=DataType.INT64, description="primary field", is_primary=True, auto_id=True
    )
    embeddings_field = FieldSchema(name="embeddings", dtype=DataType.FLOAT_VECTOR, dim=384)
    record_id_field = FieldSchema(name="record_id", dtype=DataType.VARCHAR, max_length=64)

    schema = CollectionSchema(
        enable_dynamic_field=True,
        fields=[
            id_field,
            record_id_field,
            embeddings_field,
        ],
    )

    return schema


def get_index_params() -> IndexParams:
    index_params = IndexParams()
    index_params.add_index(field_name="embeddings", index_type="AUTOINDEX", metric_type="COSINE")
    index_params.add_index(field_name="record_id", index_type="Trie")
    return index_params


# NOTE: Precheck tests are read-only so they don't interfere with destination test,
# using scope="module" we can limit number of times the docker-compose has to be run
@pytest.fixture(scope="module")
def collection():
    docker_client = docker.from_env()
    with docker_compose_context(docker_compose_path=env_setup_path / "milvus"):
        milvus_container = docker_client.containers.get("milvus-standalone")
        healthcheck_wait(container=milvus_container)
        milvus_client = MilvusClient(uri=DB_URI)
        try:
            # Create the database
            database_resp = milvus_client._get_connection().create_database(db_name=DB_NAME)
            milvus_client.using_database(db_name=DB_NAME)

            print(f"Created database {DB_NAME}: {database_resp}")

            # Create the collection
            schema = get_schema()
            index_params = get_index_params()
            collection_resp = milvus_client.create_collection(
                collection_name=EXISTENT_COLLECTION_NAME, schema=schema, index_params=index_params
            )
            print(f"Created collection {EXISTENT_COLLECTION_NAME}: {collection_resp}")
            yield EXISTENT_COLLECTION_NAME
        finally:
            milvus_client.close()


def get_count(client: MilvusClient) -> int:
    count_field = "count(*)"
    resp = client.query(collection_name="test_collection", output_fields=[count_field])
    return resp[0][count_field]


def validate_count(
    client: MilvusClient, expected_count: int, retries: int = 10, interval: int = 1
) -> None:
    current_count = get_count(client=client)
    retry_count = 0
    while current_count != expected_count and retry_count < retries:
        time.sleep(interval)
        current_count = get_count(client=client)
        retry_count += 1
    assert current_count == expected_count, (
        f"Expected count ({expected_count}) doesn't match how "
        f"much came back from collection: {current_count}"
    )


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
async def test_milvus_destination(
    upload_file: Path,
    collection: str,
    tmp_path: Path,
):
    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=CONNECTOR_TYPE,
        identifier="mock file data",
    )
    stager = MilvusUploadStager()
    uploader = MilvusUploader(
        connection_config=MilvusConnectionConfig(uri=DB_URI),
        upload_config=MilvusUploaderConfig(collection_name=collection, db_name=DB_NAME),
    )
    staged_filepath = stager.run(
        elements_filepath=upload_file,
        file_data=file_data,
        output_dir=tmp_path,
        output_filename=upload_file.name,
    )
    uploader.precheck()
    uploader.run(path=staged_filepath, file_data=file_data)

    # Run validation
    with staged_filepath.open() as f:
        staged_elements = json.load(f)
    expected_count = len(staged_elements)
    with uploader.get_client() as client:
        validate_count(client=client, expected_count=expected_count)

    # Rerun and make sure the same documents get updated
    uploader.run(path=staged_filepath, file_data=file_data)
    with uploader.get_client() as client:
        validate_count(client=client, expected_count=expected_count)


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
def test_precheck_succeeds(collection: str):
    uploader = MilvusUploader(
        connection_config=MilvusConnectionConfig(uri=DB_URI),
        upload_config=MilvusUploaderConfig(db_name=DB_NAME, collection_name=collection),
    )
    uploader.precheck()


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
def test_precheck_fails_on_nonexistent_collection(collection: str):
    uploader = MilvusUploader(
        connection_config=MilvusConnectionConfig(uri=DB_URI),
        upload_config=MilvusUploaderConfig(
            db_name=DB_NAME, collection_name=NONEXISTENT_COLLECTION_NAME
        ),
    )
    with pytest.raises(
        DestinationConnectionError,
        match=f"Collection '{NONEXISTENT_COLLECTION_NAME}' does not exist",
    ):
        uploader.precheck()


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
def test_precheck_fails_on_nonexisting_db(collection: str):
    uploader = MilvusUploader(
        connection_config=MilvusConnectionConfig(uri=DB_URI),
        upload_config=MilvusUploaderConfig(db_name="nonexisting_db", collection_name=collection),
    )
    with pytest.raises(
        DestinationConnectionError,
        match="database not found",
    ):
        uploader.precheck()


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
@pytest.mark.parametrize("upload_file_str", ["upload_file_ndjson", "upload_file"])
def test_milvus_stager(
    request: TopRequest,
    upload_file_str: str,
    tmp_path: Path,
):
    upload_file: Path = request.getfixturevalue(upload_file_str)
    stager = MilvusUploadStager()
    stager_validation(
        configs=StagerValidationConfigs(test_id=CONNECTOR_TYPE, expected_count=22),
        input_file=upload_file,
        stager=stager,
        tmp_dir=tmp_path,
    )



# https://developers.notion.com/reference/block#column-list-and-column
from dataclasses import dataclass
from typing import Optional

from htmlBuilder.tags import HtmlTag

from unstructured_ingest.connector.notion.interfaces import BlockBase


@dataclass
class ColumnList(BlockBase):
    @staticmethod
    def can_have_children() -> bool:
        return True

    @classmethod
    def from_dict(cls, data: dict):
        return cls()

    def get_html(self) -> Optional[HtmlTag]:
        return None


@dataclass
class Column(BlockBase):
    @staticmethod
    def can_have_children() -> bool:
        return True

    @classmethod
    def from_dict(cls, data: dict):
        return cls()

    def get_html(self) -> Optional[HtmlTag]:
        return None



from dataclasses import dataclass
from typing import TYPE_CHECKING

from pydantic import Field, SecretStr

from unstructured_ingest.embed.interfaces import (
    AsyncBaseEmbeddingEncoder,
    BaseEmbeddingEncoder,
    EmbeddingConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.errors import (
    ProviderError,
    QuotaError,
    RateLimitError,
    UserAuthError,
    UserError,
)

if TYPE_CHECKING:
    from openai import AsyncOpenAI, OpenAI


class OctoAiEmbeddingConfig(EmbeddingConfig):
    api_key: SecretStr
    embedder_model_name: str = Field(default="thenlper/gte-large", alias="model_name")
    base_url: str = Field(default="https://text.octoai.run/v1")

    def wrap_error(self, e: Exception) -> Exception:
        # https://platform.openai.com/docs/guides/error-codes/api-errors
        from openai import APIStatusError

        if not isinstance(e, APIStatusError):
            logger.error(f"unhandled exception from openai: {e}", exc_info=True)
            raise e
        error_code = e.code
        if 400 <= e.status_code < 500:
            # user error
            if e.status_code == 401:
                return UserAuthError(e.message)
            if e.status_code == 429:
                # 429 indicates rate limit exceeded and quote exceeded
                if error_code == "insufficient_quota":
                    return QuotaError(e.message)
                else:
                    return RateLimitError(e.message)
            return UserError(e.message)
        if e.status_code >= 500:
            return ProviderError(e.message)
        logger.error(f"unhandled exception from openai: {e}", exc_info=True)
        return e

    @requires_dependencies(
        ["openai", "tiktoken"],
        extras="embed-octoai",
    )
    def get_client(self) -> "OpenAI":
        """Creates an OpenAI python client to embed elements. Uses the OpenAI SDK."""
        from openai import OpenAI

        return OpenAI(api_key=self.api_key.get_secret_value(), base_url=self.base_url)

    @requires_dependencies(
        ["openai", "tiktoken"],
        extras="embed-octoai",
    )
    def get_async_client(self) -> "AsyncOpenAI":
        """Creates an OpenAI python client to embed elements. Uses the OpenAI SDK."""
        from openai import AsyncOpenAI

        return AsyncOpenAI(api_key=self.api_key.get_secret_value(), base_url=self.base_url)


@dataclass
class OctoAIEmbeddingEncoder(BaseEmbeddingEncoder):
    config: OctoAiEmbeddingConfig

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    def embed_query(self, query: str):
        try:
            client = self.config.get_client()
            response = client.embeddings.create(input=query, model=self.config.embedder_model_name)
        except Exception as e:
            raise self.wrap_error(e=e)
        return response.data[0].embedding

    def embed_documents(self, elements: list[dict]) -> list[dict]:
        texts = [e.get("text", "") for e in elements]
        try:
            client = self.config.get_client()
            response = client.embeddings.create(input=texts, model=self.config.embedder_model_name)
        except Exception as e:
            raise self.wrap_error(e=e)
        embeddings = [data.embedding for data in response.data]
        elements_with_embeddings = self._add_embeddings_to_elements(elements, embeddings)
        return elements_with_embeddings


@dataclass
class AsyncOctoAIEmbeddingEncoder(AsyncBaseEmbeddingEncoder):
    config: OctoAiEmbeddingConfig

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    async def embed_query(self, query: str):
        client = self.config.get_async_client()
        try:
            response = await client.embeddings.create(
                input=query, model=self.config.embedder_model_name
            )
        except Exception as e:
            raise self.wrap_error(e=e)
        return response.data[0].embedding

    async def embed_documents(self, elements: list[dict]) -> list[dict]:
        texts = [e.get("text", "") for e in elements]
        client = self.config.get_async_client()
        try:
            response = await client.embeddings.create(
                input=texts, model=self.config.embedder_model_name
            )
        except Exception as e:
            raise self.wrap_error(e=e)
        embeddings = [data.embedding for data in response.data]
        elements_with_embeddings = self._add_embeddings_to_elements(elements, embeddings)
        return elements_with_embeddings



from dataclasses import dataclass

from unstructured_ingest.interfaces import PermissionsCleanupMixin, ProcessorConfig
from unstructured_ingest.pipeline.interfaces import PermissionsNode


@dataclass
class PermissionsDataCleaner(PermissionsNode, PermissionsCleanupMixin):
    processor_config: ProcessorConfig

    def run(self):
        self.cleanup_permissions()



from dataclasses import dataclass
from typing import TYPE_CHECKING

from pydantic import Field, SecretStr

from unstructured_ingest.embed.interfaces import (
    AsyncBaseEmbeddingEncoder,
    BaseEmbeddingEncoder,
    EmbeddingConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.errors import (
    RateLimitError as CustomRateLimitError,
)
from unstructured_ingest.v2.errors import (
    UserAuthError,
    UserError,
)

if TYPE_CHECKING:
    from together import AsyncTogether, Together


class TogetherAIEmbeddingConfig(EmbeddingConfig):
    api_key: SecretStr
    embedder_model_name: str = Field(
        default="togethercomputer/m2-bert-80M-8k-retrieval", alias="model_name"
    )

    def wrap_error(self, e: Exception) -> Exception:
        # https://docs.together.ai/docs/error-codes
        from together.error import AuthenticationError, RateLimitError, TogetherException

        if not isinstance(e, TogetherException):
            logger.error(f"unhandled exception from openai: {e}", exc_info=True)
            return e
        message = e.args[0]
        if isinstance(e, AuthenticationError):
            return UserAuthError(message)
        if isinstance(e, RateLimitError):
            return CustomRateLimitError(message)
        return UserError(message)

    @requires_dependencies(["together"], extras="togetherai")
    def get_client(self) -> "Together":
        from together import Together

        return Together(api_key=self.api_key.get_secret_value())

    @requires_dependencies(["together"], extras="togetherai")
    def get_async_client(self) -> "AsyncTogether":
        from together import AsyncTogether

        return AsyncTogether(api_key=self.api_key.get_secret_value())


@dataclass
class TogetherAIEmbeddingEncoder(BaseEmbeddingEncoder):
    config: TogetherAIEmbeddingConfig

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    def embed_query(self, query: str) -> list[float]:
        return self._embed_documents(elements=[query])[0]

    def embed_documents(self, elements: list[dict]) -> list[dict]:
        embeddings = self._embed_documents([e.get("text", "") for e in elements])
        return self._add_embeddings_to_elements(elements, embeddings)

    def _embed_documents(self, elements: list[str]) -> list[list[float]]:
        client = self.config.get_client()
        try:
            outputs = client.embeddings.create(
                model=self.config.embedder_model_name, input=elements
            )
        except Exception as e:
            raise self.wrap_error(e=e)
        return [outputs.data[i].embedding for i in range(len(elements))]


@dataclass
class AsyncTogetherAIEmbeddingEncoder(AsyncBaseEmbeddingEncoder):
    config: TogetherAIEmbeddingConfig

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    async def embed_query(self, query: str) -> list[float]:
        embedding = await self._embed_documents(elements=[query])
        return embedding[0]

    async def embed_documents(self, elements: list[dict]) -> list[dict]:
        embeddings = await self._embed_documents([e.get("text", "") for e in elements])
        return self._add_embeddings_to_elements(elements, embeddings)

    async def _embed_documents(self, elements: list[str]) -> list[list[float]]:
        client = self.config.get_async_client()
        try:
            outputs = await client.embeddings.create(
                model=self.config.embedder_model_name, input=elements
            )
        except Exception as e:
            raise self.wrap_error(e=e)
        return [outputs.data[i].embedding for i in range(len(elements))]



from abc import ABC
from dataclasses import dataclass
from pathlib import Path
from typing import Any, TypeVar

from pydantic import BaseModel

from unstructured_ingest.utils.data_prep import get_data
from unstructured_ingest.v2.interfaces.connector import BaseConnector
from unstructured_ingest.v2.interfaces.file_data import FileData
from unstructured_ingest.v2.interfaces.process import BaseProcess


class UploaderConfig(BaseModel):
    pass


UploaderConfigT = TypeVar("UploaderConfigT", bound=UploaderConfig)


@dataclass
class UploadContent:
    path: Path
    file_data: FileData


@dataclass
class Uploader(BaseProcess, BaseConnector, ABC):
    upload_config: UploaderConfigT
    connector_type: str

    def is_async(self) -> bool:
        return False

    def is_batch(self) -> bool:
        return False

    def run_batch(self, contents: list[UploadContent], **kwargs: Any) -> None:
        raise NotImplementedError()

    def run(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        data = get_data(path=path)
        self.run_data(data=data, file_data=file_data, **kwargs)

    async def run_async(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        data = get_data(path=path)
        await self.run_data_async(data=data, file_data=file_data, **kwargs)

    def run_data(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        raise NotImplementedError()

    async def run_data_async(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        return self.run_data(data=data, file_data=file_data, **kwargs)



import hashlib
import time
from dataclasses import dataclass, field
from datetime import timezone
from pathlib import Path
from typing import TYPE_CHECKING, Any, Coroutine, Generator

from pydantic import Field, Secret

from unstructured_ingest.error import SourceConnectionError
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    FileData,
    Indexer,
    IndexerConfig,
)
from unstructured_ingest.v2.interfaces.file_data import FileDataSourceMetadata, SourceIdentifiers
from unstructured_ingest.v2.processes.connector_registry import SourceRegistryEntry

MAX_EMAILS_PER_FOLDER = 1_000_000  # Maximum number of emails per folder

if TYPE_CHECKING:
    from office365.graph_client import GraphClient
    from office365.outlook.mail.folders.folder import MailFolder
    from office365.outlook.mail.messages.message import Message


CONNECTOR_TYPE = "outlook"


class OutlookAccessConfig(AccessConfig):
    client_credential: str = Field(description="Azure AD App client secret", alias="client_cred")


class OutlookConnectionConfig(ConnectionConfig):
    access_config: Secret[OutlookAccessConfig]
    client_id: str = Field(description="Azure AD App client ID")
    tenant: str = Field(
        default="common", description="ID or domain name associated with your Azure AD instance"
    )
    authority_url: str = Field(
        default="https://login.microsoftonline.com",
        description="Authentication token provider for Microsoft apps",
    )

    @requires_dependencies(["msal"], extras="outlook")
    def _acquire_token(self):
        """Acquire token via MSAL"""
        from msal import ConfidentialClientApplication

        # NOTE: It'd be nice to use `msal.authority.AuthorityBuilder` here paired with AZURE_PUBLIC
        # constant as default in the future but they do not fit well with `authority_url` right now
        authority_url = f"{self.authority_url.rstrip('/')}/{self.tenant}"
        app = ConfidentialClientApplication(
            authority=authority_url,
            client_id=self.client_id,
            client_credential=self.access_config.get_secret_value().client_credential,
        )
        token = app.acquire_token_for_client(scopes=["https://graph.microsoft.com/.default"])
        return token

    @requires_dependencies(["office365"], extras="outlook")
    @SourceConnectionError.wrap
    def get_client(self) -> "GraphClient":
        from office365.graph_client import GraphClient

        return GraphClient(self._acquire_token)


class OutlookIndexerConfig(IndexerConfig):
    outlook_folders: list[str] = Field(
        description="Folders to download email messages from. Do not specify subfolders. "
        "Use quotes if there are spaces in folder names."
    )
    recursive: bool = Field(
        default=False,
        description="Recursively download files in their respective folders otherwise stop at the"
        " files in provided folder level.",
    )
    user_email: str = Field(description="Outlook email to download messages from.")


@dataclass
class OutlookIndexer(Indexer):
    index_config: OutlookIndexerConfig
    connection_config: OutlookConnectionConfig
    connector_type: str = CONNECTOR_TYPE

    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        messages = self._list_messages(recursive=self.index_config.recursive)

        for message in messages:
            yield self._message_to_file_data(message)

    def run_async(self, **kwargs: Any) -> Coroutine[Any, Any, Any]:
        raise NotImplementedError

    @SourceConnectionError.wrap
    def precheck(self) -> None:
        client = self.connection_config.get_client()
        client.users[self.index_config.user_email].get().execute_query()

    def is_async(self) -> bool:
        return False

    def _list_messages(self, recursive: bool) -> list["Message"]:
        mail_folders = self._get_selected_root_folders()
        messages = []

        while mail_folders:
            mail_folder = mail_folders.pop()
            messages += list(mail_folder.messages.get().top(MAX_EMAILS_PER_FOLDER).execute_query())

            if recursive:
                mail_folders += list(mail_folder.child_folders.get().execute_query())

        return messages

    def _get_selected_root_folders(self) -> list["MailFolder"]:
        client_user = self.connection_config.get_client().users[self.index_config.user_email]
        root_mail_folders = client_user.mail_folders.get().execute_query()

        selected_names_normalized = [
            folder_name.lower() for folder_name in self.index_config.outlook_folders
        ]
        selected_root_mail_folders = [
            folder
            for folder in root_mail_folders
            if folder.display_name.lower() in selected_names_normalized
        ]

        if not selected_root_mail_folders:
            logger.error(
                f"Root folders selected in configuration: {self.index_config.outlook_folders}"
                f"not found for user email {self.index_config.user_email}. Aborting."
            )
            raise ValueError("Root folders selected in configuration not found.")

        return selected_root_mail_folders

    def _message_to_file_data(self, message: "Message") -> FileData:
        fullpath = self._generate_fullpath(message)

        return FileData(
            identifier=message.id,
            connector_type=CONNECTOR_TYPE,
            source_identifiers=SourceIdentifiers(filename=fullpath.name, fullpath=str(fullpath)),
            metadata=FileDataSourceMetadata(
                url=message.resource_url,
                version=message.change_key,
                date_modified=str(
                    message.last_modified_datetime.replace(tzinfo=timezone.utc).timestamp()
                ),
                date_created=str(message.created_datetime.replace(tzinfo=timezone.utc).timestamp()),
                date_processed=str(time.time()),
                record_locator={
                    "message_id": message.id,
                    "user_email": self.index_config.user_email,
                },
            ),
            additional_metadata={
                "sent_from": str(message.sent_from),
                "to_recipients": [str(recipient) for recipient in message.to_recipients],
                "bcc_recipients": [str(recipient) for recipient in message.to_recipients],
                "subject": message.subject,
                "conversation_id": message.conversation_id,
                "is_draft": message.is_draft,
                "is_read": message.is_read,
                "has_attachments": message.has_attachments,
                "importance": message.importance,
            },
        )

    def _generate_fullpath(self, message: "Message") -> Path:
        return Path(hashlib.sha256(message.id.encode("utf-8")).hexdigest()[:16] + ".eml")


class OutlookDownloaderConfig(DownloaderConfig):
    pass


@dataclass
class OutlookDownloader(Downloader):
    connector_type: str = CONNECTOR_TYPE
    connection_config: OutlookConnectionConfig
    download_config: OutlookDownloaderConfig = field(default_factory=OutlookDownloaderConfig)

    def run(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        # NOTE: Indexer should provide source identifiers required to generate the download path
        download_path = self.get_download_path(file_data)
        if download_path is None:
            logger.error(
                "Generated download path is None, source_identifiers might be missing"
                "from FileData."
            )
            raise ValueError("Generated invalid download path.")

        self._download_message(file_data, download_path)
        return self.generate_download_response(file_data, download_path)

    def is_async(self) -> bool:
        return False

    def _download_message(self, file_data: FileData, download_path: Path) -> None:
        # NOTE: Indexer should supply the record locator in metadata
        if (
            file_data.metadata.record_locator is None
            or "user_email" not in file_data.metadata.record_locator
            or "message_id" not in file_data.metadata.record_locator
        ):
            logger.error(
                f"Invalid record locator in metadata: {file_data.metadata.record_locator}."
                "Keys 'user_email' and 'message_id' must be present."
            )
            raise ValueError("Invalid record locator.")

        user_email = file_data.metadata.record_locator["user_email"]
        message_id = file_data.metadata.record_locator["message_id"]

        message = self.connection_config.get_client().users[user_email].messages[message_id]
        download_path.parent.mkdir(exist_ok=True, parents=True)

        with open(download_path, "wb") as file:
            message.download(file).execute_query()


outlook_source_entry = SourceRegistryEntry(
    indexer=OutlookIndexer,
    indexer_config=OutlookIndexerConfig,
    downloader=OutlookDownloader,
    downloader_config=OutlookDownloaderConfig,
    connection_config=OutlookConnectionConfig,
)



import csv
import hashlib
from dataclasses import dataclass, field
from pathlib import Path
from time import time
from typing import TYPE_CHECKING, Any, Generator, Optional

from pydantic import BaseModel, Field, Secret

from unstructured_ingest import __name__ as integration_name
from unstructured_ingest.__version__ import __version__ as integration_version
from unstructured_ingest.error import (
    DestinationConnectionError,
    SourceConnectionError,
    SourceConnectionNetworkError,
)
from unstructured_ingest.utils.data_prep import batch_generator, get_data
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.utils.string_and_date_utils import truncate_string_bytes
from unstructured_ingest.v2.constants import RECORD_ID_LABEL
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    BatchFileData,
    BatchItem,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    FileData,
    FileDataSourceMetadata,
    Indexer,
    IndexerConfig,
    SourceIdentifiers,
    Uploader,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
    download_responses,
)
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)

if TYPE_CHECKING:
    from astrapy import AsyncCollection as AstraDBAsyncCollection
    from astrapy import Collection as AstraDBCollection
    from astrapy import DataAPIClient as AstraDBClient


CONNECTOR_TYPE = "astradb"

MAX_CONTENT_PARAM_BYTE_SIZE = 8000


class AstraDBAdditionalMetadata(BaseModel):
    collection_name: str
    keyspace: Optional[str] = None


class AstraDBBatchFileData(BatchFileData):
    additional_metadata: AstraDBAdditionalMetadata


class AstraDBAccessConfig(AccessConfig):
    token: str = Field(description="Astra DB Token with access to the database.")
    api_endpoint: str = Field(description="The API endpoint for the Astra DB.")


class AstraDBConnectionConfig(ConnectionConfig):
    connector_type: str = Field(default=CONNECTOR_TYPE, init=False)
    access_config: Secret[AstraDBAccessConfig]

    @requires_dependencies(["astrapy"], extras="astradb")
    def get_client(self) -> "AstraDBClient":
        from astrapy import DataAPIClient as AstraDBClient

        # Create a client object to interact with the Astra DB
        # caller_name/version for Astra DB tracking
        return AstraDBClient(
            caller_name=integration_name,
            caller_version=integration_version,
        )


def get_astra_collection(
    connection_config: AstraDBConnectionConfig,
    collection_name: str,
    keyspace: str,
) -> "AstraDBCollection":
    # Build the Astra DB object.
    access_configs = connection_config.access_config.get_secret_value()

    # Create a client object to interact with the Astra DB
    # caller_name/version for Astra DB tracking
    client = connection_config.get_client()

    # Get the database object
    astra_db = client.get_database(
        api_endpoint=access_configs.api_endpoint,
        token=access_configs.token,
        keyspace=keyspace,
    )

    # Connect to the collection
    astra_db_collection = astra_db.get_collection(name=collection_name)
    return astra_db_collection


async def get_async_astra_collection(
    connection_config: AstraDBConnectionConfig,
    collection_name: str,
    keyspace: str,
) -> "AstraDBAsyncCollection":
    # Build the Astra DB object.
    access_configs = connection_config.access_config.get_secret_value()

    # Create a client object to interact with the Astra DB
    client = connection_config.get_client()

    # Get the async database object
    async_astra_db = client.get_async_database(
        api_endpoint=access_configs.api_endpoint,
        token=access_configs.token,
        keyspace=keyspace,
    )

    # Get async collection from AsyncDatabase
    async_astra_db_collection = await async_astra_db.get_collection(name=collection_name)
    return async_astra_db_collection


class AstraDBUploadStagerConfig(UploadStagerConfig):
    pass


class AstraDBIndexerConfig(IndexerConfig):
    collection_name: str = Field(
        description="The name of the Astra DB collection. "
        "Note that the collection name must only include letters, "
        "numbers, and underscores."
    )
    keyspace: Optional[str] = Field(default=None, description="The Astra DB connection keyspace.")
    batch_size: int = Field(default=20, description="Number of records per batch")


class AstraDBDownloaderConfig(DownloaderConfig):
    fields: list[str] = field(default_factory=list)


class AstraDBUploaderConfig(UploaderConfig):
    collection_name: str = Field(
        description="The name of the Astra DB collection. "
        "Note that the collection name must only include letters, "
        "numbers, and underscores."
    )
    keyspace: Optional[str] = Field(default=None, description="The Astra DB connection keyspace.")
    requested_indexing_policy: Optional[dict[str, Any]] = Field(
        default=None,
        description="The indexing policy to use for the collection.",
        examples=['{"deny": ["metadata"]}'],
    )
    batch_size: int = Field(default=20, description="Number of records per batch")
    record_id_key: str = Field(
        default=RECORD_ID_LABEL,
        description="searchable key to find entries for the same record on previous runs",
    )


@dataclass
class AstraDBIndexer(Indexer):
    connection_config: AstraDBConnectionConfig
    index_config: AstraDBIndexerConfig

    def get_collection(self) -> "AstraDBCollection":
        return get_astra_collection(
            connection_config=self.connection_config,
            collection_name=self.index_config.collection_name,
            keyspace=self.index_config.keyspace,
        )

    def precheck(self) -> None:
        try:
            self.get_collection().options()
        except Exception as e:
            logger.error(f"Failed to validate connection {e}", exc_info=True)
            raise SourceConnectionError(f"failed to validate connection: {e}")

    def _get_doc_ids(self) -> set[str]:
        """Fetches all document ids in an index"""
        # Get the collection
        collection = self.get_collection()

        # Perform the find operation to get all items
        astra_db_docs_cursor = collection.find({}, projection={"_id": True})

        # Iterate over the cursor
        astra_db_docs = []
        for result in astra_db_docs_cursor:
            astra_db_docs.append(result)

        # Create file data for each astra record
        ids = sorted([astra_record["_id"] for astra_record in astra_db_docs])

        return set(ids)

    def run(self, **kwargs: Any) -> Generator[AstraDBBatchFileData, None, None]:
        all_ids = self._get_doc_ids()
        ids = list(all_ids)
        id_batches = batch_generator(ids, self.index_config.batch_size)

        for batch in id_batches:
            fd = AstraDBBatchFileData(
                connector_type=CONNECTOR_TYPE,
                metadata=FileDataSourceMetadata(
                    date_processed=str(time()),
                ),
                additional_metadata=AstraDBAdditionalMetadata(
                    collection_name=self.index_config.collection_name,
                    keyspace=self.index_config.keyspace,
                ),
                batch_items=[BatchItem(identifier=b) for b in batch],
            )
            yield fd


@dataclass
class AstraDBDownloader(Downloader):
    connection_config: AstraDBConnectionConfig
    download_config: AstraDBDownloaderConfig
    connector_type: str = CONNECTOR_TYPE

    def is_async(self) -> bool:
        return True

    def get_identifier(self, record_id: str) -> str:
        f = f"{record_id}"
        if self.download_config.fields:
            f = "{}-{}".format(
                f,
                hashlib.sha256(",".join(self.download_config.fields).encode()).hexdigest()[:8],
            )
        return f

    def write_astra_result_to_csv(self, astra_result: dict, download_path: str) -> None:
        with open(download_path, "w", encoding="utf8") as f:
            writer = csv.writer(f)
            writer.writerow(astra_result.keys())
            writer.writerow(astra_result.values())

    def generate_download_response(
        self, result: dict, file_data: AstraDBBatchFileData
    ) -> DownloadResponse:
        record_id = result["_id"]
        filename_id = self.get_identifier(record_id=record_id)
        filename = f"{filename_id}.csv"  # csv to preserve column info
        download_path = self.download_dir / Path(filename)
        logger.debug(f"Downloading results from record {record_id} as csv to {download_path}")
        download_path.parent.mkdir(parents=True, exist_ok=True)
        try:
            self.write_astra_result_to_csv(astra_result=result, download_path=str(download_path))
        except Exception as e:
            logger.error(
                f"failed to download from record {record_id} to {download_path}: {e}",
                exc_info=True,
            )
            raise SourceConnectionNetworkError(f"failed to download file {file_data.identifier}")

        # modify input file_data for download_response
        file_data.source_identifiers = SourceIdentifiers(filename=filename, fullpath=filename)
        cast_file_data = FileData.cast(file_data=file_data)
        cast_file_data.identifier = filename
        cast_file_data.metadata.date_processed = str(time())
        cast_file_data.metadata.record_locator = {"document_id": record_id}
        return super().generate_download_response(
            file_data=cast_file_data, download_path=download_path
        )

    def run(self, file_data: FileData, **kwargs: Any) -> download_responses:
        raise NotImplementedError("Use astradb run_async instead")

    async def run_async(self, file_data: FileData, **kwargs: Any) -> download_responses:
        # Get metadata from file_data
        astra_file_data = AstraDBBatchFileData.cast(file_data=file_data)
        ids: list[str] = [item.identifier for item in astra_file_data.batch_items]
        collection_name: str = astra_file_data.additional_metadata.collection_name
        keyspace: str = astra_file_data.additional_metadata.keyspace

        # Retrieve results from async collection
        download_responses = []
        async_astra_collection = await get_async_astra_collection(
            connection_config=self.connection_config,
            collection_name=collection_name,
            keyspace=keyspace,
        )
        async for result in async_astra_collection.find({"_id": {"$in": ids}}):
            download_responses.append(
                self.generate_download_response(result=result, file_data=astra_file_data)
            )
        return download_responses


@dataclass
class AstraDBUploadStager(UploadStager):
    upload_stager_config: AstraDBUploadStagerConfig = field(
        default_factory=lambda: AstraDBUploadStagerConfig()
    )

    def truncate_dict_elements(self, element_dict: dict) -> None:
        text = element_dict.pop("text", None)
        if text is not None:
            element_dict["text"] = truncate_string_bytes(text, MAX_CONTENT_PARAM_BYTE_SIZE)
        metadata = element_dict.get("metadata")
        if metadata is not None and isinstance(metadata, dict):
            text_as_html = element_dict["metadata"].pop("text_as_html", None)
            if text_as_html is not None:
                element_dict["metadata"]["text_as_html"] = truncate_string_bytes(
                    text_as_html, MAX_CONTENT_PARAM_BYTE_SIZE
                )

    def conform_dict(self, element_dict: dict, file_data: FileData) -> dict:
        self.truncate_dict_elements(element_dict)
        return {
            "$vector": element_dict.pop("embeddings", None),
            "content": element_dict.pop("text", None),
            RECORD_ID_LABEL: file_data.identifier,
            "metadata": element_dict,
        }


@dataclass
class AstraDBUploader(Uploader):
    connection_config: AstraDBConnectionConfig
    upload_config: AstraDBUploaderConfig
    connector_type: str = CONNECTOR_TYPE

    def precheck(self) -> None:
        try:
            get_astra_collection(
                connection_config=self.connection_config,
                collection_name=self.upload_config.collection_name,
                keyspace=self.upload_config.keyspace,
            ).options()
        except Exception as e:
            logger.error(f"Failed to validate connection {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    @requires_dependencies(["astrapy"], extras="astradb")
    def get_collection(self) -> "AstraDBCollection":
        return get_astra_collection(
            connection_config=self.connection_config,
            collection_name=self.upload_config.collection_name,
            keyspace=self.upload_config.keyspace,
        )

    def delete_by_record_id(self, collection: "AstraDBCollection", file_data: FileData):
        logger.debug(
            f"deleting records from collection {collection.name} "
            f"with {self.upload_config.record_id_key} "
            f"set to {file_data.identifier}"
        )
        delete_filter = {self.upload_config.record_id_key: {"$eq": file_data.identifier}}
        delete_resp = collection.delete_many(filter=delete_filter)
        logger.debug(
            f"deleted {delete_resp.deleted_count} records from collection {collection.name}"
        )

    def run_data(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        logger.info(
            f"writing {len(data)} objects to destination "
            f"collection {self.upload_config.collection_name}"
        )

        astra_db_batch_size = self.upload_config.batch_size
        collection = self.get_collection()

        self.delete_by_record_id(collection=collection, file_data=file_data)

        for chunk in batch_generator(data, astra_db_batch_size):
            collection.insert_many(chunk)

    def run(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        data = get_data(path=path)
        self.run_data(data=data, file_data=file_data, **kwargs)


astra_db_source_entry = SourceRegistryEntry(
    indexer=AstraDBIndexer,
    indexer_config=AstraDBIndexerConfig,
    downloader=AstraDBDownloader,
    downloader_config=AstraDBDownloaderConfig,
    connection_config=AstraDBConnectionConfig,
)

astra_db_destination_entry = DestinationRegistryEntry(
    connection_config=AstraDBConnectionConfig,
    upload_stager_config=AstraDBUploadStagerConfig,
    upload_stager=AstraDBUploadStager,
    uploader_config=AstraDBUploaderConfig,
    uploader=AstraDBUploader,
)



# https://developers.notion.com/reference/rich-text
from dataclasses import dataclass
from typing import Optional

from htmlBuilder.attributes import Href, Style
from htmlBuilder.tags import A, B, Code, Div, HtmlTag, I, S, Span, U
from htmlBuilder.tags import Text as HtmlText

from unstructured_ingest.connector.notion.interfaces import (
    FromJSONMixin,
    GetHTMLMixin,
)
from unstructured_ingest.connector.notion.types.date import Date
from unstructured_ingest.connector.notion.types.user import People


@dataclass
class Annotations(FromJSONMixin):
    bold: bool
    code: bool
    italic: bool
    strikethrough: bool
    underline: bool
    color: str

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class Equation(FromJSONMixin, GetHTMLMixin):
    expression: str

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def get_html(self) -> Optional[HtmlTag]:
        return Code([], self.expression) if self.expression else None


@dataclass
class MentionDatabase(FromJSONMixin, GetHTMLMixin):
    id: str

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def get_html(self) -> Optional[HtmlTag]:
        return Div([], self.id) if self.id else None


@dataclass
class MentionLinkPreview(FromJSONMixin, GetHTMLMixin):
    url: str

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def get_html(self) -> Optional[HtmlTag]:
        return A([Href(self.url)], self.url) if self.url else None


@dataclass
class MentionPage(FromJSONMixin, GetHTMLMixin):
    id: str

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def get_html(self) -> Optional[HtmlTag]:
        return Div([], self.id) if self.id else None


@dataclass
class MentionTemplate(FromJSONMixin):
    template_mention_date: Optional[str]
    template_mention_user: Optional[str]

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class Mention(FromJSONMixin, GetHTMLMixin):
    type: str
    database: Optional[MentionDatabase] = None
    date: Optional[Date] = None
    link_preview: Optional[MentionLinkPreview] = None
    page: Optional[MentionPage] = None
    template_mention: Optional[MentionTemplate] = None
    user: Optional[People] = None

    @classmethod
    def from_dict(cls, data: dict):
        t = data["type"]
        mention = cls(type=t)
        if t == "date":
            mention.date = Date.from_dict(data["date"])
        elif t == "database":
            mention.database = MentionDatabase.from_dict(data["database"])
        elif t == "link_preview":
            mention.link_preview = MentionLinkPreview.from_dict(data["link_preview"])
        elif t == "page":
            mention.page = MentionPage.from_dict(data["page"])
        elif t == "template_mention":
            mention.template_mention = MentionTemplate.from_dict(data["template_mention"])
        elif t == "user":
            mention.user = People.from_dict(data["user"])

        return mention

    def get_html(self) -> Optional[HtmlTag]:
        t = self.type
        if t == "date":
            return self.date.get_html() if self.date else None
        elif t == "database":
            return self.database.get_html() if self.database else None
        elif t == "link_preview":
            return self.link_preview.get_html() if self.link_preview else None
        elif t == "page":
            return self.page.get_html() if self.page else None
        elif t == "user":
            return self.user.get_html() if self.user else None
        return None


@dataclass
class Text(FromJSONMixin):
    content: str
    link: Optional[dict]

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class RichText(FromJSONMixin, GetHTMLMixin):
    type: str
    plain_text: str
    annotations: Optional[Annotations] = None
    href: Optional[str] = None
    text: Optional[Text] = None
    mention: Optional[Mention] = None
    equation: Optional[Equation] = None

    def get_html(self) -> Optional[HtmlTag]:
        text = HtmlText(self.plain_text)
        if self.href:
            text = A([Href(self.href)], text)
        if self.annotations:
            annotations = self.annotations
            if annotations.bold:
                text = B([], text)
            if annotations.code:
                text = Code([], text)
            if annotations.italic:
                text = I([], text)
            if annotations.strikethrough:
                text = S([], text)
            if annotations.underline:
                text = U([], text)
            if annotations.color and annotations.color != "default":
                if isinstance(text, HtmlText):
                    text = Span([], text)
                text.attributes.append(Style(f"color:{annotations.color}"))
        return text

    @classmethod
    def from_dict(cls, data: dict):
        t = data["type"]
        rich_text = cls(
            annotations=Annotations.from_dict(data.pop("annotations")),
            **data,
        )
        if t == "text":
            rich_text.text = Text.from_dict(data["text"])
        elif t == "mention":
            rich_text.mention = Mention.from_dict(data["mention"])
        elif t == "equation":
            rich_text.equation = Equation.from_dict(data["equation"])

        return rich_text



from __future__ import annotations

import asyncio
import logging
import multiprocessing as mp
import shutil
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Awaitable, Callable, Optional, TypeVar

from tqdm import tqdm
from tqdm.asyncio import tqdm as tqdm_asyncio

from unstructured_ingest.v2.interfaces import BaseProcess, ProcessorConfig, Uploader
from unstructured_ingest.v2.logger import logger, make_default_logger
from unstructured_ingest.v2.otel import OtelHandler
from unstructured_ingest.v2.pipeline.otel import instrument

BaseProcessT = TypeVar("BaseProcessT", bound=BaseProcess)
iterable_input = list[dict[str, Any]]


@dataclass
class PipelineStep(ABC):
    process: BaseProcessT
    context: ProcessorConfig
    identifier: str

    def __str__(self):
        return self.identifier

    def process_serially(self, iterable: iterable_input) -> Any:
        logger.info("processing content serially")
        if iterable:
            if len(iterable) == 1:
                return [self.run(**iterable[0])]
            if self.context.tqdm:
                return [self.run(**it) for it in tqdm(iterable, desc=self.identifier)]
            return [self.run(**it) for it in iterable]
        return [self.run()]

    async def _process_async(self, iterable: iterable_input) -> Any:
        if iterable:
            if len(iterable) == 1:
                return [await self.run_async(**iterable[0])]
            if self.context.tqdm:
                return await tqdm_asyncio.gather(
                    *[self.run_async(**i) for i in iterable], desc=self.identifier
                )
            return await asyncio.gather(*[self.run_async(**i) for i in iterable])
        return [await self.run_async()]

    def process_async(self, iterable: iterable_input) -> Any:
        logger.info("processing content async")
        return self.asyncio_run(fn=self._process_async, iterable=iterable)

    def asyncio_run(
        self, fn: Callable[[Any, Any], Awaitable[Any]], *args: Any, **kwargs: Any
    ) -> Any:
        current_loop = asyncio._get_running_loop()
        if current_loop is None:
            return asyncio.run(fn(*args, **kwargs))
        with ThreadPoolExecutor(thread_name_prefix="asyncio") as thread_pool:
            logger.warning(
                f"async code being run in dedicated thread pool "
                f"to not conflict with existing event loop: {current_loop}"
            )

            def wrapped():
                return asyncio.run(fn(*args, **kwargs))

            future = thread_pool.submit(wrapped)
            return future.result()

    def process_multiprocess(self, iterable: iterable_input) -> Any:
        logger.info("processing content across processes")

        if iterable:
            if len(iterable) == 1:
                return self.process_serially(iterable)
            if self.context.num_processes == 1:
                return self.process_serially(iterable)
            with mp.Pool(
                processes=self.context.num_processes,
                initializer=self._init_mp,
                initargs=(
                    logging.DEBUG if self.context.verbose else logging.INFO,
                    self.context.otel_endpoint,
                ),
            ) as pool:
                otel_context = OtelHandler.inject_context()
                for iter in iterable:
                    iter[OtelHandler.trace_context_key] = otel_context
                if self.context.tqdm:
                    return list(
                        tqdm(
                            pool.imap_unordered(func=self._wrap_mp, iterable=iterable),
                            total=len(iterable),
                            desc=self.identifier,
                        )
                    )
                return pool.map(self._wrap_mp, iterable)
        return [self.run()]

    def _wrap_mp(self, input_kwargs: dict) -> Any:
        # Allow mapping of kwargs via multiprocessing map()
        return self.run(**input_kwargs)

    def _init_mp(self, log_level: int, endpoint: Optional[str] = None) -> None:
        # Init logger for each spawned process when using multiprocessing pool
        make_default_logger(level=log_level)
        otel_handler = OtelHandler(otel_endpoint=endpoint, log_out=logger.debug)
        otel_handler.init_trace()

    @instrument()
    def __call__(self, iterable: Optional[iterable_input] = None) -> Any:
        iterable = iterable or []
        if iterable:
            logger.info(
                f"calling {self.__class__.__name__} " f"with {len(iterable)} docs",  # type: ignore
            )
        else:
            logger.info(f"calling {self.__class__.__name__} with no inputs")
        if self.context.async_supported and self.process.is_async():
            return self.process_async(iterable=iterable)
        if self.context.mp_supported:
            return self.process_multiprocess(iterable=iterable)
        return self.process_serially(iterable=iterable)

    def _run(self, fn: Callable, **kwargs: Any) -> Optional[Any]:
        return self.asyncio_run(fn=self.run_async, _fn=fn, **kwargs)

    async def _run_async(self, fn: Callable, **kwargs: Any) -> Optional[Any]:
        raise NotImplementedError

    def run(self, _fn: Callable[..., Any] | None = None, **kwargs: Any) -> Optional[Any]:
        kwargs = kwargs.copy()
        otel_handler = OtelHandler(otel_endpoint=self.context.otel_endpoint, log_out=logger.debug)
        tracer = otel_handler.get_tracer()
        if trace_context := kwargs.pop(otel_handler.trace_context_key, {}):
            otel_handler.attach_context(trace_context=trace_context)
        attributes = {}
        if file_data_path := kwargs.get("file_data_path"):
            attributes["file_id"] = Path(file_data_path).stem
        try:
            with tracer.start_as_current_span(self.identifier, record_exception=True) as span:
                otel_handler.set_attributes(span, attributes)
                fn = _fn or self.process.run
                return self._run(fn=fn, **kwargs)
        except Exception as e:
            logger.error(f"Exception raised while running {self.identifier}", exc_info=e)
            if "file_data_path" in kwargs:
                self.context.status[kwargs["file_data_path"]] = {self.identifier: str(e)}
            if self.context.raise_on_error:
                raise e
            return None

    async def run_async(self, _fn: Optional[Callable] = None, **kwargs: Any) -> Optional[Any]:
        otel_handler = OtelHandler(otel_endpoint=self.context.otel_endpoint, log_out=logger.debug)
        try:
            attributes = {}
            if file_data_path := kwargs.get("file_data_path"):
                attributes["file_id"] = Path(file_data_path).stem
            with otel_handler.get_tracer().start_as_current_span(
                self.identifier, record_exception=True
            ) as span:
                otel_handler.set_attributes(span, attributes)
                fn = _fn or self.process.run_async
                return await self._run_async(fn=fn, **kwargs)
        except Exception as e:
            logger.error(f"Exception raised while running {self.identifier}", exc_info=e)
            if "file_data_path" in kwargs:
                self.context.status[kwargs["file_data_path"]] = {self.identifier: str(e)}
            if self.context.raise_on_error:
                raise e
            return None

    @property
    def cache_dir(self) -> Path:
        return Path(self.context.work_dir) / self.identifier

    def delete_cache(self):
        if self.context.iter_delete and self.cache_dir.exists():
            cache_dir = self.cache_dir
            logger.info(f"deleting {self.identifier} cache dir {cache_dir}")
            shutil.rmtree(cache_dir)


@dataclass
class BatchPipelineStep(PipelineStep, ABC):
    process: Uploader

    def __call__(self, iterable: Optional[iterable_input] = None) -> Any:
        if self.context.mp_supported and self.process.is_batch():
            return self.run_batch(contents=iterable)
        super().__call__(iterable=iterable)

    @abstractmethod
    def _run_batch(self, contents: iterable_input, **kwargs) -> Any:
        pass

    def run_batch(self, contents: iterable_input, **kwargs) -> Any:
        try:
            return self._run_batch(contents=contents, **kwargs)
        except Exception as e:
            self.context.status[self.identifier] = {"step_error": str(e)}
            if self.context.raise_on_error:
                raise e
            return None



import copy
import typing as t
from dataclasses import dataclass, field
from pathlib import Path

from unstructured_ingest import __name__ as integration_name
from unstructured_ingest.__version__ import __version__ as integration_version
from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.enhanced_dataclass.core import _asdict
from unstructured_ingest.error import DestinationConnectionError, SourceConnectionError
from unstructured_ingest.interfaces import (
    AccessConfig,
    BaseConnectorConfig,
    BaseDestinationConnector,
    BaseSingleIngestDoc,
    BaseSourceConnector,
    IngestDocCleanupMixin,
    SourceConnectorCleanupMixin,
    SourceMetadata,
    WriteConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.data_prep import batch_generator, flatten_dict
from unstructured_ingest.utils.dep_check import requires_dependencies

if t.TYPE_CHECKING:
    from astrapy import Collection as AstraDBCollection
    from astrapy import Database as AstraDB

NON_INDEXED_FIELDS = ["metadata._node_content", "content"]


@dataclass
class AstraDBAccessConfig(AccessConfig):
    token: str = enhanced_field(sensitive=True)
    api_endpoint: str = enhanced_field(sensitive=True)


@dataclass
class SimpleAstraDBConfig(BaseConnectorConfig):
    access_config: AstraDBAccessConfig
    collection_name: str
    keyspace: t.Optional[str] = None
    namespace: t.Optional[str] = None


@dataclass
class AstraDBIngestDoc(IngestDocCleanupMixin, BaseSingleIngestDoc):
    connector_config: SimpleAstraDBConfig
    metadata: t.Dict[str, str] = field(default_factory=dict)
    registry_name: str = "astradb"

    @property
    def filename(self):
        return (
            Path(self.read_config.download_dir)
            / self.connector_config.collection_name
            / f"{self.metadata['_id']}.txt"
        ).resolve()

    @property
    def _output_filename(self):
        return (
            Path(self.processor_config.output_dir)
            / self.connector_config.collection_name
            / f"{self.metadata['_id']}.json"
        ).resolve()

    def update_source_metadata(self, **kwargs):
        if not self.metadata:
            self.source_metadata = SourceMetadata(
                exists=False,
            )
            return
        self.source_metadata = SourceMetadata(
            exists=True,
        )

    @SourceConnectionError.wrap
    @requires_dependencies(["astrapy"], extras="astradb")
    @BaseSingleIngestDoc.skip_if_file_exists
    def get_file(self):
        self.filename.parent.mkdir(parents=True, exist_ok=True)

        flattened_dict = flatten_dict(dictionary=self.metadata)
        str_values = [str(value) for value in flattened_dict.values()]
        concatenated_values = "\n".join(str_values)

        with open(self.filename, "w") as f:
            f.write(concatenated_values)


@dataclass
class AstraDBSourceConnector(SourceConnectorCleanupMixin, BaseSourceConnector):
    connector_config: SimpleAstraDBConfig
    _astra_db: t.Optional["AstraDB"] = field(init=False, default=None)
    _astra_db_collection: t.Optional["AstraDBCollection"] = field(init=False, default=None)

    @property
    @requires_dependencies(["astrapy"], extras="astradb")
    def astra_db_collection(self) -> "AstraDBCollection":
        if self._astra_db_collection is None:
            from astrapy import DataAPIClient as AstraDBClient

            # Choose keyspace or deprecated namespace
            keyspace_param = self.connector_config.keyspace or self.connector_config.namespace

            # Create a client object to interact with the Astra DB
            # caller_name/version for Astra DB tracking
            my_client = AstraDBClient(
                caller_name=integration_name,
                caller_version=integration_version,
            )

            # Get the database object
            self._astra_db = my_client.get_database(
                api_endpoint=self.connector_config.access_config.api_endpoint,
                token=self.connector_config.access_config.token,
                keyspace=keyspace_param,
            )

            # Create and connect to the newly created collection
            self._astra_db_collection = self._astra_db.get_collection(
                name=self.connector_config.collection_name,
            )

        return self._astra_db_collection  # type: ignore

    @requires_dependencies(["astrapy"], extras="astradb")
    @SourceConnectionError.wrap  # type: ignore
    def initialize(self):
        _ = self.astra_db_collection

    @requires_dependencies(["astrapy"], extras="astradb")
    def check_connection(self):
        try:
            _ = self.astra_db_collection
        except Exception as e:
            logger.error(f"Failed to validate connection {e}", exc_info=True)
            raise SourceConnectionError(f"failed to validate connection: {e}")

    @requires_dependencies(["astrapy"], extras="astradb")
    def get_ingest_docs(self):  # type: ignore
        # Perform the find operation
        astra_db_docs_cursor = self.astra_db_collection.find({})

        # Iterate over the cursor
        astra_db_docs = []
        for result in astra_db_docs_cursor:
            astra_db_docs.append(result)

        # Create a list of AstraDBIngestDoc objects
        doc_list = []
        for record in astra_db_docs:
            doc = AstraDBIngestDoc(
                connector_config=self.connector_config,
                processor_config=self.processor_config,
                read_config=self.read_config,
                metadata=record,
            )

            doc.update_source_metadata()

            doc_list.append(doc)

        return doc_list


@dataclass
class AstraDBWriteConfig(WriteConfig):
    embedding_dimension: int
    requested_indexing_policy: t.Optional[t.Dict[str, t.Any]] = None
    batch_size: int = 20


@dataclass
class AstraDBDestinationConnector(BaseDestinationConnector):
    write_config: AstraDBWriteConfig
    connector_config: SimpleAstraDBConfig
    _astra_db: t.Optional["AstraDB"] = field(init=False, default=None)
    _astra_db_collection: t.Optional["AstraDBCollection"] = field(init=False, default=None)

    def to_dict(self, **kwargs):
        """
        The _astra_db_collection variable in this dataclass breaks deepcopy due to:
        TypeError: cannot pickle '_thread.lock' object
        When serializing, remove it, meaning client data will need to be reinitialized
        when deserialized
        """
        self_cp = copy.copy(self)

        if hasattr(self_cp, "_astra_db_collection"):
            setattr(self_cp, "_astra_db_collection", None)

        return _asdict(self_cp, **kwargs)

    @property
    @requires_dependencies(["astrapy"], extras="astradb")
    def astra_db_collection(self) -> "AstraDBCollection":
        if self._astra_db_collection is None:
            from astrapy import DataAPIClient as AstraDBClient
            from astrapy.exceptions import CollectionAlreadyExistsException

            # Choose keyspace or deprecated namespace
            keyspace_param = self.connector_config.keyspace or self.connector_config.namespace

            collection_name = self.connector_config.collection_name
            embedding_dimension = self.write_config.embedding_dimension
            requested_indexing_policy = self.write_config.requested_indexing_policy

            # Create a client object to interact with the Astra DB
            # caller_name/version for Astra DB tracking
            my_client = AstraDBClient(
                caller_name=integration_name,
                caller_version=integration_version,
            )

            # Get the database object
            self._astra_db = my_client.get_database(
                api_endpoint=self.connector_config.access_config.api_endpoint,
                token=self.connector_config.access_config.token,
                keyspace=keyspace_param,
            )

            # Create and connect to the newly created collection
            try:
                self._astra_db_collection = self._astra_db.create_collection(
                    name=collection_name,
                    dimension=embedding_dimension,
                    indexing=requested_indexing_policy,
                )
            except CollectionAlreadyExistsException as e:
                logger.info(f"{e}", exc_info=True)
                self._astra_db_collection = self._astra_db.get_collection(name=collection_name)

        return self._astra_db_collection

    @requires_dependencies(["astrapy"], extras="astradb")
    @DestinationConnectionError.wrap
    def initialize(self):
        _ = self.astra_db_collection

    @requires_dependencies(["astrapy"], extras="astradb")
    def check_connection(self):
        try:
            _ = self.astra_db_collection
        except Exception as e:
            logger.error(f"Failed to validate connection {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    def write_dict(self, *args, elements_dict: t.List[t.Dict[str, t.Any]], **kwargs) -> None:
        logger.info(f"inserting / updating {len(elements_dict)} documents to Astra DB.")

        if self._astra_db_collection is None:
            raise DestinationConnectionError("Astra DB collection not available for insertion.")

        astra_db_batch_size = self.write_config.batch_size

        for batch in batch_generator(elements_dict, astra_db_batch_size):
            self._astra_db_collection.insert_many(batch)

    def normalize_dict(self, element_dict: dict) -> dict:
        return {
            "$vector": element_dict.pop("embeddings", None),
            "content": element_dict.pop("text", None),
            "metadata": element_dict,
        }



import filecmp
import shutil
from pathlib import Path
from typing import Callable, Optional

from pydantic import BaseModel

from test.integration.connectors.utils.constants import expected_results_path
from test.integration.connectors.utils.validation.equality import file_type_equality_check


class ValidationConfig(BaseModel):
    test_id: str
    file_equality_check: Optional[Callable[[Path, Path], bool]] = None

    def test_output_dir(self) -> Path:
        return expected_results_path / self.test_id

    def detect_diff(self, expected_filepath: Path, current_filepath: Path) -> bool:
        if expected_filepath.suffix != current_filepath.suffix:
            return True
        if file_equality_check := self.file_equality_check:
            return not file_equality_check(expected_filepath, current_filepath)
        current_suffix = expected_filepath.suffix
        if current_suffix in file_type_equality_check:
            equality_check_callable = file_type_equality_check[current_suffix]
            return not equality_check_callable(
                expected_filepath=expected_filepath, current_filepath=current_filepath
            )
        # Fallback is using filecmp.cmp to compare the files
        return not filecmp.cmp(expected_filepath, current_filepath, shallow=False)


def reset_dir(dir_path: Path) -> None:
    shutil.rmtree(path=dir_path, ignore_errors=True)
    dir_path.mkdir(parents=True)



import asyncio
from abc import ABC, abstractmethod
from dataclasses import dataclass

import numpy as np
from pydantic import BaseModel


class EmbeddingConfig(BaseModel):
    pass


@dataclass
class BaseEncoder(ABC):
    config: EmbeddingConfig

    def initialize(self):
        """Initializes the embedding encoder class. Should also validate the instance
        is properly configured: e.g., embed a single a element"""

    def wrap_error(self, e: Exception) -> Exception:
        """Handle errors from the embedding service. Should raise a more informative error
        if possible"""
        return e

    @staticmethod
    def _add_embeddings_to_elements(
        elements: list[dict], embeddings: list[list[float]]
    ) -> list[dict]:
        """
        Add embeddings to elements.

        Args:
            elements (list[Element]): List of elements.
            embeddings (list[list[float]]): List of embeddings.

        Returns:
            list[Element]: Elements with embeddings added.
        """
        assert len(elements) == len(embeddings)
        elements_w_embedding = []
        for i, element in enumerate(elements):
            element["embeddings"] = embeddings[i]
            elements_w_embedding.append(element)
        return elements


@dataclass
class BaseEmbeddingEncoder(BaseEncoder, ABC):

    def initialize(self):
        """Initializes the embedding encoder class. Should also validate the instance
        is properly configured: e.g., embed a single a element"""

    @property
    def num_of_dimensions(self) -> tuple[int, ...]:
        exemplary_embedding = self.get_exemplary_embedding()
        return np.shape(exemplary_embedding)

    def get_exemplary_embedding(self) -> list[float]:
        return self.embed_query(query="Q")

    @property
    def is_unit_vector(self) -> bool:
        """Denotes if the embedding vector is a unit vector."""
        exemplary_embedding = self.get_exemplary_embedding()
        return np.isclose(np.linalg.norm(exemplary_embedding), 1.0)

    @abstractmethod
    def embed_documents(self, elements: list[dict]) -> list[dict]:
        pass

    @abstractmethod
    def embed_query(self, query: str) -> list[float]:
        pass

    def _embed_documents(self, elements: list[str]) -> list[list[float]]:
        results = []
        for text in elements:
            response = self.embed_query(query=text)
            results.append(response)

        return results


@dataclass
class AsyncBaseEmbeddingEncoder(BaseEncoder, ABC):

    async def initialize(self):
        """Initializes the embedding encoder class. Should also validate the instance
        is properly configured: e.g., embed a single a element"""

    @property
    async def num_of_dimensions(self) -> tuple[int, ...]:
        exemplary_embedding = await self.get_exemplary_embedding()
        return np.shape(exemplary_embedding)

    async def get_exemplary_embedding(self) -> list[float]:
        return await self.embed_query(query="Q")

    @property
    async def is_unit_vector(self) -> bool:
        """Denotes if the embedding vector is a unit vector."""
        exemplary_embedding = await self.get_exemplary_embedding()
        return np.isclose(np.linalg.norm(exemplary_embedding), 1.0)

    @abstractmethod
    async def embed_documents(self, elements: list[dict]) -> list[dict]:
        pass

    @abstractmethod
    async def embed_query(self, query: str) -> list[float]:
        pass

    async def _embed_documents(self, elements: list[str]) -> list[list[float]]:
        results = await asyncio.gather(*[self.embed_query(query=text) for text in elements])
        return results



class UserError(Exception):
    pass


class UserAuthError(UserError):
    pass


class RateLimitError(UserError):
    pass


class QuotaError(UserError):
    pass


class ProviderError(Exception):
    pass



import json
import os
from pathlib import Path

import pytest

from test.integration.embedders.utils import (
    validate_embedding_output,
    validate_raw_embedder,
    validate_raw_embedder_async,
)
from test.integration.utils import requires_env
from unstructured_ingest.embed.mixedbreadai import (
    AsyncMixedbreadAIEmbeddingEncoder,
    MixedbreadAIEmbeddingConfig,
    MixedbreadAIEmbeddingEncoder,
)
from unstructured_ingest.v2.processes.embedder import Embedder, EmbedderConfig

API_KEY = "MXBAI_API_KEY"


def get_api_key() -> str:
    api_key = os.getenv(API_KEY, None)
    assert api_key
    return api_key


@requires_env(API_KEY)
def test_mixedbread_embedder(embedder_file: Path):
    api_key = get_api_key()
    embedder_config = EmbedderConfig(embedding_provider="mixedbread-ai", embedding_api_key=api_key)
    embedder = Embedder(config=embedder_config)
    results = embedder.run(elements_filepath=embedder_file)
    assert results
    with embedder_file.open("r") as f:
        original_elements = json.load(f)
    validate_embedding_output(original_elements=original_elements, output_elements=results)


@requires_env(API_KEY)
def test_raw_mixedbread_embedder(embedder_file: Path):
    api_key = get_api_key()
    embedder = MixedbreadAIEmbeddingEncoder(
        config=MixedbreadAIEmbeddingConfig(
            api_key=api_key,
        )
    )
    validate_raw_embedder(
        embedder=embedder,
        embedder_file=embedder_file,
        expected_dimensions=(1024,),
        expected_is_unit_vector=False,
    )


@requires_env(API_KEY)
@pytest.mark.asyncio
async def test_raw_async_mixedbread_embedder(embedder_file: Path):
    api_key = get_api_key()
    embedder = AsyncMixedbreadAIEmbeddingEncoder(
        config=MixedbreadAIEmbeddingConfig(
            api_key=api_key,
        )
    )
    await validate_raw_embedder_async(
        embedder=embedder,
        embedder_file=embedder_file,
        expected_dimensions=(1024,),
        expected_is_unit_vector=False,
    )



import typing as t
from dataclasses import dataclass

from unstructured_ingest.interfaces import BaseDestinationConnector
from unstructured_ingest.runner.writers.base_writer import Writer

if t.TYPE_CHECKING:
    from unstructured_ingest.connector.fsspec.azure import (
        AzureWriteConfig,
        SimpleAzureBlobStorageConfig,
    )


@dataclass
class AzureWriter(Writer):
    connector_config: "SimpleAzureBlobStorageConfig"
    write_config: "AzureWriteConfig"

    def get_connector_cls(self) -> t.Type[BaseDestinationConnector]:
        from unstructured_ingest.connector.fsspec.azure import (
            AzureBlobStorageDestinationConnector,
        )

        return AzureBlobStorageDestinationConnector



import typing as t
from dataclasses import dataclass

from unstructured_ingest.interfaces import BaseDestinationConnector
from unstructured_ingest.runner.writers.base_writer import Writer

if t.TYPE_CHECKING:
    from unstructured_ingest.connector.sql import SimpleSqlConfig
    from unstructured_ingest.interfaces import WriteConfig


@dataclass
class SqlWriter(Writer):
    write_config: "WriteConfig"
    connector_config: "SimpleSqlConfig"

    def get_connector_cls(self) -> t.Type[BaseDestinationConnector]:
        from unstructured_ingest.connector.sql import (
            SqlDestinationConnector,
        )

        return SqlDestinationConnector



import json
import os
from pathlib import Path

import pytest

from test.integration.utils import requires_env
from unstructured_ingest.v2.errors import UserError
from unstructured_ingest.v2.processes.partitioner import Partitioner, PartitionerConfig

int_test_dir = Path(__file__).parent
assets_dir = int_test_dir / "assets"

all_partition_files = [path for path in assets_dir.iterdir() if path.is_file()]
non_image_partition_files = [
    path for path in all_partition_files if path.suffix not in [".jpg", ".png", ".tif"]
]
image_partition_files = [
    path for path in all_partition_files if path not in non_image_partition_files
]


@pytest.mark.parametrize(
    "partition_file", all_partition_files, ids=[path.name for path in all_partition_files]
)
@requires_env("UNSTRUCTURED_API_KEY", "UNSTRUCTURED_API_URL")
@pytest.mark.asyncio
async def test_partitioner_api_hi_res(partition_file: Path):
    api_key = os.getenv("UNSTRUCTURED_API_KEY")
    api_url = os.getenv("UNSTRUCTURED_API_URL")
    partitioner_config = PartitionerConfig(
        strategy="hi_res", partition_by_api=True, api_key=api_key, partition_endpoint=api_url
    )
    partitioner = Partitioner(config=partitioner_config)
    results = await partitioner.run_async(filename=partition_file)
    results_dir = int_test_dir / "results"
    results_dir.mkdir(exist_ok=True)
    results_path = results_dir / f"{partition_file.name}.json"
    with results_path.open("w") as f:
        json.dump(results, f, indent=2)
    assert results


@pytest.mark.parametrize(
    "partition_file",
    non_image_partition_files,
    ids=[path.name for path in non_image_partition_files],
)
@requires_env("UNSTRUCTURED_API_KEY", "UNSTRUCTURED_API_URL")
@pytest.mark.asyncio
async def test_partitioner_api_fast(partition_file: Path):
    api_key = os.getenv("UNSTRUCTURED_API_KEY")
    api_url = os.getenv("UNSTRUCTURED_API_URL")
    partitioner_config = PartitionerConfig(
        strategy="fast", partition_by_api=True, api_key=api_key, partition_endpoint=api_url
    )
    partitioner = Partitioner(config=partitioner_config)
    results = await partitioner.run_async(filename=partition_file)
    assert results


@pytest.mark.parametrize(
    "partition_file", image_partition_files, ids=[path.name for path in image_partition_files]
)
@requires_env("UNSTRUCTURED_API_KEY", "UNSTRUCTURED_API_URL")
@pytest.mark.asyncio
async def test_partitioner_api_fast_error(partition_file: Path):
    api_key = os.getenv("UNSTRUCTURED_API_KEY")
    api_url = os.getenv("UNSTRUCTURED_API_URL")
    partitioner_config = PartitionerConfig(
        strategy="fast", partition_by_api=True, api_key=api_key, partition_endpoint=api_url
    )
    partitioner = Partitioner(config=partitioner_config)
    with pytest.raises(UserError):
        await partitioner.run_async(filename=partition_file)



from __future__ import annotations

import asyncio
import json
import uuid
from collections import defaultdict
from contextlib import asynccontextmanager
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any, AsyncGenerator, Optional

from pydantic import BaseModel, ConfigDict, Field, Secret

from unstructured_ingest.error import DestinationConnectionError
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.chunking import elements_from_base64_gzipped_json
from unstructured_ingest.utils.data_prep import batch_generator
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    FileData,
    Uploader,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
)
from unstructured_ingest.v2.processes.connector_registry import (
    DestinationRegistryEntry,
)

if TYPE_CHECKING:
    from neo4j import AsyncDriver, Auth
    from networkx import Graph, MultiDiGraph

CONNECTOR_TYPE = "neo4j"


class Neo4jAccessConfig(AccessConfig):
    password: str


class Neo4jConnectionConfig(ConnectionConfig):
    access_config: Secret[Neo4jAccessConfig]
    connector_type: str = Field(default=CONNECTOR_TYPE, init=False)
    username: str
    uri: str = Field(description="Neo4j Connection URI <scheme>://<host>:<port>")
    database: str = Field(description="Name of the target database")

    @requires_dependencies(["neo4j"], extras="neo4j")
    @asynccontextmanager
    async def get_client(self) -> AsyncGenerator["AsyncDriver", None]:
        from neo4j import AsyncGraphDatabase

        driver = AsyncGraphDatabase.driver(**self._get_driver_parameters())
        logger.info(f"Created driver connecting to the database '{self.database}' at {self.uri}.")
        try:
            yield driver
        finally:
            await driver.close()
            logger.info(
                f"Closed driver connecting to the database '{self.database}' at {self.uri}."
            )

    def _get_driver_parameters(self) -> dict:
        return {
            "uri": self.uri,
            "auth": self._get_auth(),
            "database": self.database,
        }

    @requires_dependencies(["neo4j"], extras="neo4j")
    def _get_auth(self) -> "Auth":
        from neo4j import Auth

        return Auth("basic", self.username, self.access_config.get_secret_value().password)


class Neo4jUploadStagerConfig(UploadStagerConfig):
    pass


@dataclass
class Neo4jUploadStager(UploadStager):
    upload_stager_config: Neo4jUploadStagerConfig = Field(
        default_factory=Neo4jUploadStagerConfig, validate_default=True
    )

    def run(  # type: ignore
        self,
        elements_filepath: Path,
        file_data: FileData,
        output_dir: Path,
        output_filename: str,
        **kwargs: Any,
    ) -> Path:
        with elements_filepath.open() as file:
            elements = json.load(file)

        nx_graph = self._create_lexical_graph(
            elements, self._create_document_node(file_data=file_data)
        )
        output_filepath = Path(output_dir) / f"{output_filename}.json"
        output_filepath.parent.mkdir(parents=True, exist_ok=True)

        with open(output_filepath, "w") as file:
            file.write(_GraphData.from_nx(nx_graph).model_dump_json())

        return output_filepath

    def _create_lexical_graph(self, elements: list[dict], document_node: _Node) -> "Graph":
        import networkx as nx

        graph = nx.MultiDiGraph()
        graph.add_node(document_node)

        previous_node: Optional[_Node] = None
        for element in elements:
            element_node = self._create_element_node(element)
            order_relationship = (
                Relationship.NEXT_CHUNK if self._is_chunk(element) else Relationship.NEXT_ELEMENT
            )
            if previous_node:
                graph.add_edge(element_node, previous_node, relationship=order_relationship)

            previous_node = element_node
            graph.add_edge(element_node, document_node, relationship=Relationship.PART_OF_DOCUMENT)

            if self._is_chunk(element):
                origin_element_nodes = [
                    self._create_element_node(origin_element)
                    for origin_element in self._get_origin_elements(element)
                ]
                graph.add_edges_from(
                    [
                        (origin_element_node, element_node)
                        for origin_element_node in origin_element_nodes
                    ],
                    relationship=Relationship.PART_OF_CHUNK,
                )
                graph.add_edges_from(
                    [
                        (origin_element_node, document_node)
                        for origin_element_node in origin_element_nodes
                    ],
                    relationship=Relationship.PART_OF_DOCUMENT,
                )

        return graph

    # TODO(Filip Knefel): Ensure _is_chunk is as reliable as possible, consider different checks
    def _is_chunk(self, element: dict) -> bool:
        return "orig_elements" in element.get("metadata", {})

    def _create_document_node(self, file_data: FileData) -> _Node:
        properties = {}
        if file_data.source_identifiers:
            properties["name"] = file_data.source_identifiers.filename
        if file_data.metadata.date_created:
            properties["date_created"] = file_data.metadata.date_created
        if file_data.metadata.date_modified:
            properties["date_modified"] = file_data.metadata.date_modified
        return _Node(id_=file_data.identifier, properties=properties, labels=[Label.DOCUMENT])

    def _create_element_node(self, element: dict) -> _Node:
        properties = {"id": element["element_id"], "text": element["text"]}

        if embeddings := element.get("embeddings"):
            properties["embeddings"] = embeddings

        label = Label.CHUNK if self._is_chunk(element) else Label.UNSTRUCTURED_ELEMENT
        return _Node(id_=element["element_id"], properties=properties, labels=[label])

    def _get_origin_elements(self, chunk_element: dict) -> list[dict]:
        orig_elements = chunk_element.get("metadata", {}).get("orig_elements")
        return elements_from_base64_gzipped_json(raw_s=orig_elements)


class _GraphData(BaseModel):
    nodes: list[_Node]
    edges: list[_Edge]

    @classmethod
    def from_nx(cls, nx_graph: "MultiDiGraph") -> _GraphData:
        nodes = list(nx_graph.nodes())
        edges = [
            _Edge(
                source_id=u.id_,
                destination_id=v.id_,
                relationship=Relationship(data_dict["relationship"]),
            )
            for u, v, data_dict in nx_graph.edges(data=True)
        ]
        return _GraphData(nodes=nodes, edges=edges)


class _Node(BaseModel):
    model_config = ConfigDict()

    id_: str = Field(default_factory=lambda: str(uuid.uuid4()))
    labels: list[Label] = Field(default_factory=list)
    properties: dict = Field(default_factory=dict)

    def __hash__(self):
        return hash(self.id_)


class _Edge(BaseModel):
    model_config = ConfigDict()

    source_id: str
    destination_id: str
    relationship: Relationship


class Label(Enum):
    UNSTRUCTURED_ELEMENT = "UnstructuredElement"
    CHUNK = "Chunk"
    DOCUMENT = "Document"


class Relationship(Enum):
    PART_OF_DOCUMENT = "PART_OF_DOCUMENT"
    PART_OF_CHUNK = "PART_OF_CHUNK"
    NEXT_CHUNK = "NEXT_CHUNK"
    NEXT_ELEMENT = "NEXT_ELEMENT"


class Neo4jUploaderConfig(UploaderConfig):
    batch_size: int = Field(
        default=100, description="Maximal number of nodes/relationships created per transaction."
    )


@dataclass
class Neo4jUploader(Uploader):
    upload_config: Neo4jUploaderConfig
    connection_config: Neo4jConnectionConfig
    connector_type: str = CONNECTOR_TYPE

    @DestinationConnectionError.wrap
    def precheck(self) -> None:
        async def verify_auth():
            async with self.connection_config.get_client() as client:
                await client.verify_connectivity()

        asyncio.run(verify_auth())

    def is_async(self):
        return True

    async def run_async(self, path: Path, file_data: FileData, **kwargs) -> None:  # type: ignore
        with path.open() as file:
            staged_data = json.load(file)

        graph_data = _GraphData.model_validate(staged_data)
        async with self.connection_config.get_client() as client:
            await self._create_uniqueness_constraints(client)
            await self._delete_old_data_if_exists(file_data, client=client)
            await self._merge_graph(graph_data=graph_data, client=client)

    async def _create_uniqueness_constraints(self, client: AsyncDriver) -> None:
        for label in Label:
            logger.info(
                f"Adding id uniqueness constraint for nodes labeled '{label.value}'"
                " if it does not already exist."
            )
            constraint_name = f"{label.value.lower()}_id"
            await client.execute_query(
                f"""
                CREATE CONSTRAINT {constraint_name} IF NOT EXISTS
                FOR (n: {label.value}) REQUIRE n.id IS UNIQUE
                """
            )

    async def _delete_old_data_if_exists(self, file_data: FileData, client: AsyncDriver) -> None:
        logger.info(f"Deleting old data for the record '{file_data.identifier}' (if present).")
        _, summary, _ = await client.execute_query(
            f"""
            MATCH (n: {Label.DOCUMENT.value} {{id: $identifier}})
            MATCH (n)--(m: {Label.CHUNK.value}|{Label.UNSTRUCTURED_ELEMENT.value})
            DETACH DELETE m""",
            identifier=file_data.identifier,
        )
        logger.info(
            f"Deleted {summary.counters.nodes_deleted} nodes"
            f" and {summary.counters.relationships_deleted} relationships."
        )

    async def _merge_graph(self, graph_data: _GraphData, client: AsyncDriver) -> None:
        nodes_by_labels: defaultdict[tuple[Label, ...], list[_Node]] = defaultdict(list)
        for node in graph_data.nodes:
            nodes_by_labels[tuple(node.labels)].append(node)

        logger.info(f"Merging {len(graph_data.nodes)} graph nodes.")
        # NOTE: Processed in parallel as there's no overlap between accessed nodes
        await self._execute_queries(
            [
                self._create_nodes_query(nodes_batch, labels)
                for labels, nodes in nodes_by_labels.items()
                for nodes_batch in batch_generator(nodes, batch_size=self.upload_config.batch_size)
            ],
            client=client,
            in_parallel=True,
        )
        logger.info(f"Finished merging {len(graph_data.nodes)} graph nodes.")

        edges_by_relationship: defaultdict[Relationship, list[_Edge]] = defaultdict(list)
        for edge in graph_data.edges:
            edges_by_relationship[edge.relationship].append(edge)

        logger.info(f"Merging {len(graph_data.edges)} graph relationships (edges).")
        # NOTE: Processed sequentially to avoid queries locking node access to one another
        await self._execute_queries(
            [
                self._create_edges_query(edges_batch, relationship)
                for relationship, edges in edges_by_relationship.items()
                for edges_batch in batch_generator(edges, batch_size=self.upload_config.batch_size)
            ],
            client=client,
        )
        logger.info(f"Finished merging {len(graph_data.edges)} graph relationships (edges).")

    @staticmethod
    async def _execute_queries(
        queries_with_parameters: list[tuple[str, dict]],
        client: AsyncDriver,
        in_parallel: bool = False,
    ) -> None:
        if in_parallel:
            logger.info(f"Executing {len(queries_with_parameters)} queries in parallel.")
            await asyncio.gather(
                *[
                    client.execute_query(query, parameters_=parameters)
                    for query, parameters in queries_with_parameters
                ]
            )
            logger.info("Finished executing parallel queries.")
        else:
            logger.info(f"Executing {len(queries_with_parameters)} queries sequentially.")
            for i, (query, parameters) in enumerate(queries_with_parameters):
                logger.info(f"Query #{i} started.")
                await client.execute_query(query, parameters_=parameters)
                logger.info(f"Query #{i} finished.")
            logger.info(
                f"Finished executing all ({len(queries_with_parameters)}) sequential queries."
            )

    @staticmethod
    def _create_nodes_query(nodes: list[_Node], labels: tuple[Label, ...]) -> tuple[str, dict]:
        labels_string = ", ".join([label.value for label in labels])
        logger.info(f"Preparing MERGE query for {len(nodes)} nodes labeled '{labels_string}'.")
        query_string = f"""
            UNWIND $nodes AS node
            MERGE (n: {labels_string} {{id: node.id}})
            SET n += node.properties
            """
        parameters = {"nodes": [{"id": node.id_, "properties": node.properties} for node in nodes]}
        return query_string, parameters

    @staticmethod
    def _create_edges_query(edges: list[_Edge], relationship: Relationship) -> tuple[str, dict]:
        logger.info(f"Preparing MERGE query for {len(edges)} {relationship} relationships.")
        query_string = f"""
            UNWIND $edges AS edge
            MATCH (u {{id: edge.source}})
            MATCH (v {{id: edge.destination}})
            MERGE (u)-[:{relationship.value}]->(v)
            """
        parameters = {
            "edges": [
                {"source": edge.source_id, "destination": edge.destination_id} for edge in edges
            ]
        }
        return query_string, parameters


neo4j_destination_entry = DestinationRegistryEntry(
    connection_config=Neo4jConnectionConfig,
    upload_stager=Neo4jUploadStager,
    upload_stager_config=Neo4jUploadStagerConfig,
    uploader=Neo4jUploader,
    uploader_config=Neo4jUploaderConfig,
)



import json
import os
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Generator, Optional

from pydantic import Field

from unstructured_ingest.v2.interfaces import FileData, Uploader, UploaderConfig
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import (
    DestinationRegistryEntry,
)
from unstructured_ingest.v2.processes.connectors.databricks.volumes import DatabricksPathMixin
from unstructured_ingest.v2.processes.connectors.sql.databricks_delta_tables import (
    DatabrickDeltaTablesConnectionConfig,
    DatabrickDeltaTablesUploadStager,
    DatabrickDeltaTablesUploadStagerConfig,
)

CONNECTOR_TYPE = "databricks_volume_delta_tables"


class DatabricksVolumeDeltaTableUploaderConfig(UploaderConfig, DatabricksPathMixin):
    database: str = Field(description="Database name", default="default")
    table_name: str = Field(description="Table name")


@dataclass
class DatabricksVolumeDeltaTableStager(DatabrickDeltaTablesUploadStager):
    def write_output(self, output_path: Path, data: list[dict], indent: Optional[int] = 2) -> None:
        # To avoid new line issues when migrating from volumes into delta tables, omit indenting
        # and always write it as a json file
        with output_path.with_suffix(".json").open("w") as f:
            json.dump(data, f)


@dataclass
class DatabricksVolumeDeltaTableUploader(Uploader):
    connection_config: DatabrickDeltaTablesConnectionConfig
    upload_config: DatabricksVolumeDeltaTableUploaderConfig
    connector_type: str = CONNECTOR_TYPE

    def precheck(self) -> None:
        with self.connection_config.get_cursor() as cursor:
            cursor.execute("SHOW CATALOGS")
            catalogs = [r[0] for r in cursor.fetchall()]
            if self.upload_config.catalog not in catalogs:
                raise ValueError(
                    "Catalog {} not found in {}".format(
                        self.upload_config.catalog, ", ".join(catalogs)
                    )
                )
            cursor.execute(f"USE CATALOG '{self.upload_config.catalog}'")
            cursor.execute("SHOW DATABASES")
            databases = [r[0] for r in cursor.fetchall()]
            if self.upload_config.database not in databases:
                raise ValueError(
                    "Database {} not found in {}".format(
                        self.upload_config.database, ", ".join(databases)
                    )
                )
            cursor.execute("SHOW TABLES")
            table_names = [r[1] for r in cursor.fetchall()]
            if self.upload_config.table_name not in table_names:
                raise ValueError(
                    "Table {} not found in {}".format(
                        self.upload_config.table_name, ", ".join(table_names)
                    )
                )

    def get_output_path(self, file_data: FileData, suffix: str = ".json") -> str:
        filename = Path(file_data.source_identifiers.filename)
        adjusted_filename = filename if filename.suffix == suffix else f"{filename}{suffix}"
        return os.path.join(self.upload_config.path, f"{adjusted_filename}")

    @contextmanager
    def get_cursor(self, **connect_kwargs) -> Generator[Any, None, None]:
        with self.connection_config.get_cursor(**connect_kwargs) as cursor:
            cursor.execute(f"USE CATALOG '{self.upload_config.catalog}'")
            yield cursor

    def run(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        with self.get_cursor(staging_allowed_local_path=str(path.parent)) as cursor:
            catalog_path = self.get_output_path(file_data=file_data)
            logger.debug(f"uploading {path.as_posix()} to {catalog_path}")
            cursor.execute(f"PUT '{path.as_posix()}' INTO '{catalog_path}' OVERWRITE")
            logger.debug(
                f"migrating content from {catalog_path} to table {self.upload_config.table_name}"
            )
            with path.open() as f:
                data = json.load(f)
                columns = data[0].keys()
            column_str = ", ".join(columns)
            sql_statment = f"INSERT INTO `{self.upload_config.table_name}` ({column_str}) SELECT {column_str} FROM json.`{catalog_path}`"  # noqa: E501
            cursor.execute(sql_statment)


databricks_volumes_delta_tables_destination_entry = DestinationRegistryEntry(
    connection_config=DatabrickDeltaTablesConnectionConfig,
    uploader=DatabricksVolumeDeltaTableUploader,
    uploader_config=DatabricksVolumeDeltaTableUploaderConfig,
    upload_stager=DatabricksVolumeDeltaTableStager,
    upload_stager_config=DatabrickDeltaTablesUploadStagerConfig,
)



from abc import ABC
from dataclasses import dataclass
from typing import Optional, Type, TypeVar

from unstructured_ingest.v2.interfaces import (
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    Indexer,
    IndexerConfig,
    Uploader,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
)

IndexerT = TypeVar("IndexerT", bound=Indexer)
IndexerConfigT = TypeVar("IndexerConfigT", bound=IndexerConfig)
DownloaderT = TypeVar("DownloaderT", bound=Downloader)
DownloaderConfigT = TypeVar("DownloaderConfigT", bound=DownloaderConfig)
ConnectionConfigT = TypeVar("ConnectionConfigT", bound=ConnectionConfig)
UploadStagerConfigT = TypeVar("UploadStagerConfigT", bound=UploadStagerConfig)
UploadStagerT = TypeVar("UploadStagerT", bound=UploadStager)
UploaderConfigT = TypeVar("UploaderConfigT", bound=UploaderConfig)
UploaderT = TypeVar("UploaderT", bound=Uploader)


@dataclass
class RegistryEntry(ABC):
    pass


@dataclass
class SourceRegistryEntry(RegistryEntry):
    indexer: Type[IndexerT]
    downloader: Type[DownloaderT]

    downloader_config: Optional[Type[DownloaderConfigT]] = None
    indexer_config: Optional[Type[IndexerConfigT]] = None
    connection_config: Optional[Type[ConnectionConfigT]] = None


source_registry: dict[str, SourceRegistryEntry] = {}


def add_source_entry(source_type: str, entry: SourceRegistryEntry):
    if source_type in source_registry:
        raise ValueError(f"source {source_type} has already been registered")
    source_registry[source_type] = entry


@dataclass
class DestinationRegistryEntry(RegistryEntry):
    uploader: Type[UploaderT]
    upload_stager: Optional[Type[UploadStagerT]] = None

    upload_stager_config: Optional[Type[UploadStagerConfigT]] = None
    uploader_config: Optional[Type[UploaderConfigT]] = None

    connection_config: Optional[Type[ConnectionConfigT]] = None


destination_registry: dict[str, DestinationRegistryEntry] = {}


def add_destination_entry(destination_type: str, entry: DestinationRegistryEntry):
    if destination_type in destination_registry:
        raise ValueError(f"destination {destination_type} has already been registered")
    destination_registry[destination_type] = entry



import json
from typing import Dict, Type, cast

from unstructured_ingest.connector.airtable import AirtableIngestDoc
from unstructured_ingest.connector.astradb import AstraDBIngestDoc
from unstructured_ingest.connector.biomed import BiomedIngestDoc
from unstructured_ingest.connector.confluence import ConfluenceIngestDoc
from unstructured_ingest.connector.delta_table import DeltaTableIngestDoc
from unstructured_ingest.connector.discord import DiscordIngestDoc
from unstructured_ingest.connector.elasticsearch import (
    ElasticsearchIngestDoc,
    ElasticsearchIngestDocBatch,
)
from unstructured_ingest.connector.fsspec.azure import AzureBlobStorageIngestDoc
from unstructured_ingest.connector.fsspec.box import BoxIngestDoc
from unstructured_ingest.connector.fsspec.dropbox import DropboxIngestDoc
from unstructured_ingest.connector.fsspec.gcs import GcsIngestDoc
from unstructured_ingest.connector.fsspec.s3 import S3IngestDoc
from unstructured_ingest.connector.fsspec.sftp import SftpIngestDoc
from unstructured_ingest.connector.github import GitHubIngestDoc
from unstructured_ingest.connector.gitlab import GitLabIngestDoc
from unstructured_ingest.connector.google_drive import GoogleDriveIngestDoc
from unstructured_ingest.connector.hubspot import HubSpotIngestDoc
from unstructured_ingest.connector.jira import JiraIngestDoc
from unstructured_ingest.connector.kafka import KafkaIngestDoc
from unstructured_ingest.connector.local import LocalIngestDoc
from unstructured_ingest.connector.mongodb import MongoDBIngestDoc, MongoDBIngestDocBatch
from unstructured_ingest.connector.notion.connector import (
    NotionDatabaseIngestDoc,
    NotionPageIngestDoc,
)
from unstructured_ingest.connector.onedrive import OneDriveIngestDoc
from unstructured_ingest.connector.opensearch import OpenSearchIngestDoc, OpenSearchIngestDocBatch
from unstructured_ingest.connector.outlook import OutlookIngestDoc
from unstructured_ingest.connector.reddit import RedditIngestDoc
from unstructured_ingest.connector.salesforce import SalesforceIngestDoc
from unstructured_ingest.connector.sharepoint import SharepointIngestDoc
from unstructured_ingest.connector.slack import SlackIngestDoc
from unstructured_ingest.connector.wikipedia import (
    WikipediaIngestHTMLDoc,
    WikipediaIngestSummaryDoc,
    WikipediaIngestTextDoc,
)
from unstructured_ingest.enhanced_dataclass import EnhancedDataClassJsonMixin
from unstructured_ingest.interfaces import BaseIngestDoc

INGEST_DOC_NAME_TO_CLASS: Dict[str, Type[EnhancedDataClassJsonMixin]] = {
    "airtable": AirtableIngestDoc,
    "astradb": AstraDBIngestDoc,
    "azure": AzureBlobStorageIngestDoc,
    "biomed": BiomedIngestDoc,
    "box": BoxIngestDoc,
    "confluence": ConfluenceIngestDoc,
    "delta-table": DeltaTableIngestDoc,
    "discord": DiscordIngestDoc,
    "dropbox": DropboxIngestDoc,
    "elasticsearch": ElasticsearchIngestDoc,
    "elasticsearch_batch": ElasticsearchIngestDocBatch,
    "gcs": GcsIngestDoc,
    "github": GitHubIngestDoc,
    "gitlab": GitLabIngestDoc,
    "google_drive": GoogleDriveIngestDoc,
    "hubspot": HubSpotIngestDoc,
    "jira": JiraIngestDoc,
    "kafka": KafkaIngestDoc,
    "local": LocalIngestDoc,
    "mongodb": MongoDBIngestDoc,
    "mongodb_batch": MongoDBIngestDocBatch,
    "notion_database": NotionDatabaseIngestDoc,
    "notion_page": NotionPageIngestDoc,
    "onedrive": OneDriveIngestDoc,
    "opensearch": OpenSearchIngestDoc,
    "opensearch_batch": OpenSearchIngestDocBatch,
    "outlook": OutlookIngestDoc,
    "reddit": RedditIngestDoc,
    "s3": S3IngestDoc,
    "salesforce": SalesforceIngestDoc,
    "sftp": SftpIngestDoc,
    "sharepoint": SharepointIngestDoc,
    "slack": SlackIngestDoc,
    "wikipedia_html": WikipediaIngestHTMLDoc,
    "wikipedia_text": WikipediaIngestTextDoc,
    "wikipedia_summary": WikipediaIngestSummaryDoc,
}


def create_ingest_doc_from_json(ingest_doc_json: str) -> BaseIngestDoc:
    try:
        ingest_doc_dict: dict = json.loads(ingest_doc_json)
    except TypeError as te:
        raise TypeError(
            f"failed to load json string when deserializing IngestDoc: {ingest_doc_json}",
        ) from te
    return create_ingest_doc_from_dict(ingest_doc_dict)


def create_ingest_doc_from_dict(ingest_doc_dict: dict) -> BaseIngestDoc:
    ingest_doc_dict = ingest_doc_dict.copy()
    if "registry_name" not in ingest_doc_dict:
        raise ValueError(f"registry_name not present in ingest doc: {ingest_doc_dict}")
    registry_name = ingest_doc_dict.pop("registry_name")
    try:
        ingest_doc_cls = INGEST_DOC_NAME_TO_CLASS[registry_name]
        return cast(BaseIngestDoc, ingest_doc_cls.from_dict(ingest_doc_dict))
    except KeyError:
        raise ValueError(
            f"Error: Received unknown IngestDoc name: {registry_name} while deserializing",
            "IngestDoc.",
        )



import json
from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import date, datetime
from typing import TYPE_CHECKING, Any, Generator, Optional

from dateutil import parser
from pydantic import Field, Secret

from unstructured_ingest.error import DestinationConnectionError, WriteError
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.constants import RECORD_ID_LABEL
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    FileData,
    Uploader,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
)
from unstructured_ingest.v2.logger import logger

if TYPE_CHECKING:
    from weaviate.classes.init import Timeout
    from weaviate.client import WeaviateClient
    from weaviate.collections.batch.client import BatchClient

CONNECTOR_TYPE = "weaviate"


class WeaviateAccessConfig(AccessConfig, ABC):
    pass


class WeaviateConnectionConfig(ConnectionConfig, ABC):
    init_timeout: int = Field(default=2, ge=0, description="Timeout for initialization checks")
    insert_timeout: int = Field(default=90, ge=0, description="Timeout for insert operations")
    query_timeout: int = Field(default=30, ge=0, description="Timeout for query operations")
    access_config: Secret[WeaviateAccessConfig] = Field(
        default=WeaviateAccessConfig(), validate_default=True
    )

    @requires_dependencies(["weaviate"], extras="weaviate")
    def get_timeout(self) -> "Timeout":
        from weaviate.classes.init import Timeout

        return Timeout(init=self.init_timeout, query=self.query_timeout, insert=self.insert_timeout)

    @abstractmethod
    @contextmanager
    def get_client(self) -> Generator["WeaviateClient", None, None]:
        pass


class WeaviateUploadStagerConfig(UploadStagerConfig):
    pass


@dataclass
class WeaviateUploadStager(UploadStager):
    upload_stager_config: WeaviateUploadStagerConfig = field(
        default_factory=lambda: WeaviateUploadStagerConfig()
    )

    @staticmethod
    def parse_date_string(date_string: str) -> date:
        try:
            timestamp = float(date_string)
            return datetime.fromtimestamp(timestamp)
        except Exception as e:
            logger.debug(f"date {date_string} string not a timestamp: {e}")
        return parser.parse(date_string)

    def conform_dict(self, element_dict: dict, file_data: FileData) -> dict:
        """
        Updates the element dictionary to conform to the Weaviate schema
        """
        data = element_dict.copy()
        working_data = data.copy()
        # Dict as string formatting
        if (
            record_locator := working_data.get("metadata", {})
            .get("data_source", {})
            .get("record_locator")
        ):
            # Explicit casting otherwise fails schema type checking
            working_data["metadata"]["data_source"]["record_locator"] = str(
                json.dumps(record_locator)
            )

        # Array of items as string formatting
        if points := working_data.get("metadata", {}).get("coordinates", {}).get("points"):
            working_data["metadata"]["coordinates"]["points"] = str(json.dumps(points))

        if links := working_data.get("metadata", {}).get("links", {}):
            working_data["metadata"]["links"] = str(json.dumps(links))

        if permissions_data := (
            working_data.get("metadata", {}).get("data_source", {}).get("permissions_data")
        ):
            working_data["metadata"]["data_source"]["permissions_data"] = json.dumps(
                permissions_data
            )

        # Datetime formatting
        if (
            date_created := working_data.get("metadata", {})
            .get("data_source", {})
            .get("date_created")
        ):
            working_data["metadata"]["data_source"]["date_created"] = self.parse_date_string(
                date_created
            ).strftime(
                "%Y-%m-%dT%H:%M:%S.%fZ",
            )

        if (
            date_modified := working_data.get("metadata", {})
            .get("data_source", {})
            .get("date_modified")
        ):
            working_data["metadata"]["data_source"]["date_modified"] = self.parse_date_string(
                date_modified
            ).strftime(
                "%Y-%m-%dT%H:%M:%S.%fZ",
            )

        if (
            date_processed := working_data.get("metadata", {})
            .get("data_source", {})
            .get("date_processed")
        ):
            working_data["metadata"]["data_source"]["date_processed"] = self.parse_date_string(
                date_processed
            ).strftime(
                "%Y-%m-%dT%H:%M:%S.%fZ",
            )

        if last_modified := working_data.get("metadata", {}).get("last_modified"):
            working_data["metadata"]["last_modified"] = self.parse_date_string(
                last_modified
            ).strftime(
                "%Y-%m-%dT%H:%M:%S.%fZ",
            )

        # String casting
        if version := working_data.get("metadata", {}).get("data_source", {}).get("version"):
            working_data["metadata"]["data_source"]["version"] = str(version)

        if page_number := working_data.get("metadata", {}).get("page_number"):
            working_data["metadata"]["page_number"] = str(page_number)

        if regex_metadata := working_data.get("metadata", {}).get("regex_metadata"):
            working_data["metadata"]["regex_metadata"] = str(json.dumps(regex_metadata))

        working_data[RECORD_ID_LABEL] = file_data.identifier
        return working_data


class WeaviateUploaderConfig(UploaderConfig):
    collection: str = Field(description="The name of the collection this object belongs to")
    batch_size: Optional[int] = Field(default=None, description="Number of records per batch")
    requests_per_minute: Optional[int] = Field(default=None, description="Rate limit for upload")
    dynamic_batch: bool = Field(default=True, description="Whether to use dynamic batch")
    record_id_key: str = Field(
        default=RECORD_ID_LABEL,
        description="searchable key to find entries for the same record on previous runs",
    )

    def model_post_init(self, __context: Any) -> None:
        batch_types = {
            "fixed_size": self.batch_size is not None,
            "rate_limited": self.requests_per_minute is not None,
            "dynamic": self.dynamic_batch,
        }

        enabled_batch_modes = [batch_key for batch_key, flag in batch_types.items() if flag]
        if not enabled_batch_modes:
            raise ValueError("No batch mode enabled")
        if len(enabled_batch_modes) > 1:
            raise ValueError(
                "Multiple batch modes enabled, only one mode can be used: {}".format(
                    ", ".join(enabled_batch_modes)
                )
            )
        logger.info(f"Uploader config instantiated with {enabled_batch_modes[0]} batch mode")

    @contextmanager
    def get_batch_client(self, client: "WeaviateClient") -> Generator["BatchClient", None, None]:
        if self.dynamic_batch:
            with client.batch.dynamic() as batch_client:
                yield batch_client
        elif self.batch_size:
            with client.batch.fixed_size(batch_size=self.batch_size) as batch_client:
                yield batch_client
        elif self.requests_per_minute:
            with client.batch.rate_limit(
                requests_per_minute=self.requests_per_minute
            ) as batch_client:
                yield batch_client
        else:
            raise ValueError("No batch mode enabled")


@dataclass
class WeaviateUploader(Uploader, ABC):
    upload_config: WeaviateUploaderConfig
    connection_config: WeaviateConnectionConfig

    def precheck(self) -> None:
        try:
            self.connection_config.get_client()
        except Exception as e:
            logger.error(f"Failed to validate connection {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    def check_for_errors(self, client: "WeaviateClient") -> None:
        failed_uploads = client.batch.failed_objects
        if failed_uploads:
            for failure in failed_uploads:
                logger.error(
                    f"Failed to upload object with id {failure.original_uuid}: {failure.message}"
                )
            raise WriteError("Failed to upload to weaviate")

    @requires_dependencies(["weaviate"], extras="weaviate")
    def delete_by_record_id(self, client: "WeaviateClient", file_data: FileData) -> None:
        from weaviate.classes.query import Filter

        record_id = file_data.identifier
        collection = client.collections.get(self.upload_config.collection)
        delete_filter = Filter.by_property(name=self.upload_config.record_id_key).equal(
            val=record_id
        )
        # There is a configurable maximum limit (QUERY_MAXIMUM_RESULTS) on the number of
        # objects that can be deleted in a single query (default 10,000). To delete
        # more objects than the limit, re-run the query until nothing is deleted.
        while True:
            resp = collection.data.delete_many(where=delete_filter)
            if resp.failed:
                raise WriteError(
                    f"failed to delete records in collection "
                    f"{self.upload_config.collection} with record "
                    f"id property {record_id}"
                )
            if not resp.failed and not resp.successful:
                break

    def run_data(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        logger.info(
            f"writing {len(data)} objects to destination "
            f"class {self.connection_config.access_config} "
        )

        with self.connection_config.get_client() as weaviate_client:
            self.delete_by_record_id(client=weaviate_client, file_data=file_data)
            with self.upload_config.get_batch_client(client=weaviate_client) as batch_client:
                for e in data:
                    vector = e.pop("embeddings", None)
                    batch_client.add_object(
                        collection=self.upload_config.collection,
                        properties=e,
                        vector=vector,
                    )
            self.check_for_errors(client=weaviate_client)
